      PROGRAM MAIN
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
      PARAMETER (ORD2=ORDM+ORDN+3)
      PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C SAMPLE CALLING PROGRAM FOR SKAP
C THIS PROGRAM COMPUTES THE MINIMUM COST BIPARTITE MATCHING 
C OF CARDINALITY K ON A SPARSE MATRIX
C
C THE INPUT UNIT NUMBER IS ASSUMED TO BE 5
C THE OUTPUT UNIT NUMBER IS ASSUMED TO BE 6
C THE ARRAYS ARE CURRENTLY DIMENSIONED TO ALLOW INSTANCES WITH AT MOST 
C 3000 ROWS, 3000 COLUMNS AND 1000000 OF ENTRIES
C
C THE PROGRAM CAN BE TESTED ON THE INSTANCE IN FILE 'sample.dat' (THE ONE
C DESCRIBED IN DELL'AMICO, LODI, MARTELLO "EFFICIENT ALGORITHMS AND CODES
C FOR K-CARDINALITY ASSIGNMENT PROBLEMS", SECTION 3.1)
C COMPLETE DETAILS ABOUT PARAMETERS CAN BE FOUND IN THE DOCUMENTATION
C OF SUBROUTINE SKAP
C
C INPUT
C M          = NUMBER OF ROWS OF THE INSTANCE
C N          = NUMBER OF COLUMNS OF THE INSTANCE
C NE         = NUMBER OF ENTRIES OF THE INSTANCE
C K          = CARDINALITY OF THE REQUIRED ASSIGNMENT
C F_FIRST(I) = FIRST ENTRY OF F_COST(), F_HEAD() CORRESPONDING TO ROW I
C F_COST(H)  = VALUE OF THE H-TH ENTRY OF MATRIX W
C F_HEAD(H)  = COLUMN INDEX OF THE H-TH ENTRY OF MATRIX W
C B_FIRST(J) = FIRST ENTRY OF B_COST(), B_TAIL() CORRESPONDING TO COLUMN J
C B_COST(H)  = VALUE OF THE H-TH ENTRY OF MATRIX W
C B_TAIL(H)  = ROW INDEX OF THE H-TH ENTRY OF MATRIX W
C
C OUTPUT
C OPT        = 0 IF THE INSTANCE IS INFEASIBLE
C            = 1 OTHERWISE
C Z          = OPTIMAL SOLUTION VALUE 
C ROW(J)     = PREDECESSOR OF NODE J: ARC (ROW(J),J) IS CHOOSEN (J=1,...,N)
C COL(I)     = SUCCESSOR OF NODE I: ARC (I,COL(I)) IS CHOOSEN (I=1,...,M)
C
      INTEGER*4 M,N,NE,K
      INTEGER*4 F_FIRST(ORDM+1),B_FIRST(ORDN+1)
      INTEGER*4 F_COST(ORDS),B_COST(ORDS)
      INTEGER*2 F_HEAD(ORDS),B_TAIL(ORDS)
      INTEGER*4 OPT,Z
      INTEGER*4 ROW(ORDN),COL(ORDM)
C
C VARIABLES INTERNAL TO ALGORITHM SKAP
C
      INTEGER*4 MU
      INTEGER*4 U(ORDM),V(ORDN)
      INTEGER*4 F_COST2(ORDS),B_COST2(ORDS)
      INTEGER*2 B_TAIL2(ORDS),F_HEAD2(ORDS)
C
C LOCAL VARIABLES
C
      INTEGER NIN,NOUT,I,J,IN,NET
      CHARACTER*80  FNAME
C
      NIN  = 5
      NOUT = 6
      WRITE(NOUT,'('' NAME OF INPUT FILE = '',$)')
      READ (NIN,'(A)')FNAME 
C
C READ THE INSTANCE STORED IN FORWARD STAR VERSION
C
      OPEN(UNIT=10,FILE=FNAME,STATUS='OLD')
      READ(10,*)M,N,NE,K
      READ(10,*)(F_FIRST(I),I=1,M+1)
      READ(10,*)(F_HEAD(IN),IN=1,NE)
      READ(10,*)(F_COST(IN),IN=1,NE)
      CLOSE(10)
C
C BUILD THE CORRESPONDING BACKWARD STAR
C
      NET = 0
      DO J=1,N
         B_FIRST(J) = NET + 1
         DO 20 I=1,M
            DO IN=F_FIRST(I),F_FIRST(I+1)-1
               IF (F_HEAD(IN) .EQ. J) THEN
                  NET = NET + 1
                  B_TAIL(NET) = I
                  B_COST(NET) = F_COST(IN)
                  GOTO 20
               ENDIF
            ENDDO
 20      CONTINUE
      ENDDO
      B_FIRST(N+1) = NET + 1
C
C EXECUTE ALGORITHM SKAP
C
      CALL SKAP(M,N,F_FIRST,F_COST,F_HEAD,B_FIRST,B_COST,B_TAIL,K,
     +    OPT,Z,ROW,COL,U,V,MU,F_COST2,B_COST2,F_HEAD2,B_TAIL2)
C
C PRINT THE RESULT
C
      IF (OPT .EQ. 0) THEN
         WRITE(NOUT,*)' THE PROBLEM IS INFEASIBLE'
      ELSE
         WRITE(NOUT,*)' OPTIMAL SOLUTION FOUND = ',Z
      ENDIF
C
      STOP
      END
C============================================================================
      SUBROUTINE SKAP(M,N,F_FIRST,F_COST,F_HEAD,B_FIRST,B_COST,B_TAIL,K,
     +    OPT,Z,ROW,COL,U,V,MU,F_COST2,B_COST2,F_HEAD2,B_TAIL2)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
      PARAMETER (ORD2=ORDM+ORDN+3)
      PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C***************************************************************************
C COPYRIGHT DELL'AMICO-LODI-MARTELLO
C Permission to use for evaluation purposes is granted provided
C that proper acknowledgments are given.
C For a commercial use, contact alodi@deis.unibo.it
C***************************************************************************
C THIS SUBROUTINE COMPUTES THE MINIMUM COST BIPARTITE MATCHING OF
C CARDINALITY K ON A SPARSE MATRIX
C
C BASED ON THE PAPERS
C
C [1] M. DELL'AMICO, S. MARTELLO, "THE K-CARDINALITY ASSIGNMENT PROBLEM",
C     DISCR. APPL. MATH. 76, 1997, 103-121
C
C [2] M. DELL'AMICO, A. LODI, S. MARTELLO, "EFFICIENT ALGORITHMS AND CODES FOR 
C     K-CARDINALITY ASSIGNMENT PROBLEMS", **** *****
C
C THE INPUT INSTANCE MUST SATISFY THE FOLLOWING CONDITIONS:
C I)   M .LE. N
C II)  K .LE. M
C III) THE COST MATRIX MUST HAVE NONNEGATIVE ENTRIES
C***************************************************************************
C INPUT
C M    = NUMBER OF ROWS
C N    = NUMBER OF COLUMNS
C THE SPARSE MATRIX W IS STORED BOTH BY ROWS AND COLUMNS:
C THE NONNEGATIVE ENTRIES ARE STORED BY ROWS IN F_COST(), F_HEAD() 
C AND BY COLUMNS IN B_COST(), B_TAIL()
C F_FIRST(I) = FIRST ENTRY OF F_COST(), F_HEAD() CORRESPONDING TO ROW I
C F_COST(H)  = VALUE OF THE H-TH ENTRY OF MATRIX W
C F_HEAD(H)  = COLUMN INDEX OF THE H-TH ENTRY OF MATRIX W
C B_FIRST(J) = FIRST ENTRY OF B_COST(), B_TAIL() CORRESPONDING TO COLUMN J
C B_COST(H)  = VALUE OF THE H-TH ENTRY OF MATRIX W
C B_TAIL(H)  = ROW INDEX OF THE H-TH ENTRY OF MATRIX W
C K          = CARDINALITY OF THE REQUIRED MATCHING
C***************************************************************************
C OUTPUT 
C OPT        = 0 THE INSTANCE IS INFEASIBLE
C              1 THE SOLUTION IS OPTIMAL
C Z          = OPTIMAL SOLUTION VALUE
C ROW(J)     = PREDECESSOR OF NODE J: ARC (ROW(J),J) IS CHOOSEN (J=1,...,N)
C COL(I)     = SUCCESSOR OF NODE I: ARC (I,COL(I)) IS CHOOSEN (I=1,...,M)
C***************************************************************************
C INTERNAL 
C F_COST2(),B_COST2()   : INTEGER*4 DIMENSIONED AT ORDS
C F_HEAD2()             : INTEGER*2 DIMENSIONED AT ORDM  
C B_TAIL2()             : INTEGER*2 DIMENSIONED AT ORDN  
C U(I)       = DUAL VARIABLE ASSOCIATED WITH ROW  I
C V(J)       = DUAL VARIABLE ASSOCIATED WITH COLUMN J
C MU         = DUAL VARIABLE ASSOCIATED WITH THE EQUALITY CONSTRAINT
C BIGM       = COST OF A DUMMY ARC
C
      INTEGER*4 M,N,F_FIRST(ORDM+1),F_COST(ORDS),
     +          B_FIRST(ORDN+1),B_COST(ORDS)      
      INTEGER*2 F_HEAD(ORDS),B_TAIL(ORDS)
      INTEGER*4 K,ROW(ORDN),COL(ORDM),U(ORDM),V(ORDN),MU,Z,OPT
      INTEGER*4 B_COST2(ORDS),F_COST2(ORDS)
      INTEGER*2 B_TAIL2(ORDS),F_HEAD2(ORDS)
C
C LOCAL VARIABLES ************
C
      INTEGER*4 F_FIRST2(ORDM),F_LAST2(ORDM),
     +       B_FIRST2(ORDN),B_LAST2(ORDN)
      INTEGER*4 C(ORDN),R(ORDM),HROW(ORDN),HCOL(ORDM),
     +       INVR(ORDM),INVC(ORDN),
     +       SETFROW(ORDM),SETFCOL(ORDN),SETFROW2(ORDM),SETFCOL2(ORDN)
      INTEGER*4 COL2(ORDN),ROW2(ORDN),U2(ORDM),V2(ORDN)
      INTEGER*4 MINR(ORDM),MINC(ORDN),SMINR(ORDM),SMINC(ORDN),
     +       PMINR(ORDM),PMINC(ORDN),PSMINR(ORDM),PSMINC(ORDN),
     +       PTMINR(ORDM),PTMINC(ORDN),TMINR(ORDM),TMINC(ORDN)     
      INTEGER*4 Y1(ORDN),Y2(ORDN),Y3(ORDN),Y4(ORDN)
      INTEGER*4 ALFA,BETA,I,II,IX,J,JJ,JMAX,IQ,L,INF,LB,LB2,NASS,NASS1,
     +       NASS2,UBMU,IA,IMAX,HNASS,HZ,MAXROW,MAXCOL,SUMC,SUMR,
     +       LB_DELTA,SET,NUMR,NUMC,DZ,HN,
     +       MU2,Z2,HZ2,HNASS2,RID,NAF,
     +       NSETFROW,NSETFCOL,DELTA,MIN,NSETFROW2,NSETFCOL2,
     +       NACUT,ELTIM,IRETN,IB,IE,
     +       NAS,NARCS,MAXFLO,NUMAUG,NUMSTG,NNCUT,
     +       NDINIC,CK,INDEX,BIGM
      PARAMETER (BIGM = 10000000)
      LOGICAL FLAG,FROW(ORDM),FCOL(ORDN)
      LOGICAL FROW2(ORDM),FCOL2(ORDN),OK(ORDM)
C
      EQUIVALENCE (FROW(1),OK(1))
      EQUIVALENCE (MINC(1),Y1(1)),(PMINC(1),Y2(1))
      EQUIVALENCE (SMINC(1),Y3(1)),(PSMINC(1),Y4(1))
C
C VARIABLES REQUIRED BY DINIC ALGORITHM
C 
      INTEGER*2 DNFADJ(ORD4)
      INTEGER*4 DNFAPT(ORD2),DNCAP(ORD4),DNFLOW(ORD4)
      COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN05/DNFLOW
C
C******************************************************************************
C MEANING OF THE MAIN INTERNAL VARIABLES
C
C MINR(I) ( MINC(J) )   : VALUE OF THE FIRST MINIMUM ELEMENT IN ROW I (COL J)
C SMINR(I) ( SMINC(J) ) : VALUE OF THE SECOND MINIMUM ELEMENT IN ROW I (COL J)
C TMINR(I) ( TMINC(J) ) : VALUE OF THE THIRD MINIMUM ELEMENT IN ROW I (COL J)
C R(I)     : LABEL OF ROW I DURING THE SEARCH OF AN AUGMENTING PATH (I=1,...,M)
C            (DURING INITIALIZATION, I-TH ROW BY DECREASING MINIMUM OF THE ROW)
C C(J)     : LABEL OF COL J DURING THE SEARCH OF AN AUGMENTING PATH (J=1,...,N)
C            (DURING INITIALIZATION, J-TH COL BY DECREASING MINIMUM OF THE COL)
C LB_DELTA : LOWER BOUND FOR ASSIGNING K-1 ELEMENTS (USED TO REDUCE THE MATRIX)
C DELTA,DZ : UPPER BOUNDS TO THE VALUE OF AN ELEMENT TO BE INSERTED IN SOLUTION
C******************************************************************************
C
C INITIALIZATION
C         
      IF ( M .GT. N ) THEN
         WRITE(*,*)' NUMBER OF ROWS EXCEEDS NUMBER OF COLUMNS'
         RETURN
      ENDIF
      IF ( K .GT. M) THEN
         WRITE(*,*)' CARDINALITY EXCEEDS NUMBER OF ROWS'
         RETURN
      ENDIF
      INF  = 1.E9
      DO J=1,N
         V(J)      = 0
         MINC(J)   = INF
         SMINC(J)  = INF
         TMINC(J)  = INF
         PMINC(J)  = 0
         PSMINC(J) = 0
         PTMINC(J) = 0
         ROW(J)    = 0
         HROW(J)   = 0        
         V2(J)     = 0
         ROW2(J)   = 0
      ENDDO
      DO I=1,M
         U(I)      = 0
         MINR(I)   = INF
         SMINR(I)  = INF
         TMINR(I)  = INF
         PMINR(I)  = 0
         PSMINR(I) = 0
         PTMINR(I) = 0
         COL(I)    = 0
         HCOL(I)   = 0
         U2(I)     = 0
         COL2(I)   = 0
      ENDDO
      OPT    = 0
      MU     = 0
      NASS   = 0
      HNASS  = 0
      HNASS2 = 0
      Z      = 0
      HZ     = INF
      HZ2    = INF      
      SUMR   = 0
      SUMC   = 0
      IF ( M .LT. 3 ) RETURN
      CK  = M*90/100.
      NDINIC = M*N*.01
      IF ( CK .LT. K .AND. F_FIRST(M+1)-1 .LE. NDINIC ) THEN
C
C EXECUTE THE DINIC'S ALGORITHM
C E.A. DINIC, "ALGORITHMS FOR SOLUTION OF PROBLEM OF MAXIMUM FLOW IN NETWORKS 
C WITH POWER ESTIMATION", SOVIET MATHEMATICS DOKLADY 11, (1970) 1277-1280
C
         CALL DEFNET(M,N,K,F_HEAD,F_FIRST,NARCS)
         CALL DNSUB(M+N+3,NARCS,M+N+2,M+N+3,MAXFLO,
     +              NUMAUG,NUMSTG,NNCUT,NACUT,ELTIM,IRETN)
         IF ( MAXFLO .LT. K ) RETURN
         NAS = 0
         HZ  = 0
         DO 300 I=1,M
            FLAG = .FALSE.
            IB = IABS(DNFAPT(I))
            IE = IABS(DNFAPT(I+1)) - 1
            DO II=IB,IE
               IF ( DNFLOW(II) .EQ. 1 ) THEN
                  J = DNFADJ(II) - M
                  HROW(J) = I
                  DO INDEX=F_FIRST(I),F_FIRST(I+1)-1
                     IF ( F_HEAD(INDEX) .EQ. J ) THEN
                        HZ = HZ + F_COST(INDEX)
                        GOTO 290
                     ENDIF
                  ENDDO
 290              NAS = NAS + 1
                  FLAG = .TRUE.
               ENDIF
               IF ( NAS .EQ. K ) GOTO 305
               IF ( FLAG ) GOTO 300
            ENDDO
 300     CONTINUE
 305     DELTA = HZ
         DZ    = HZ
         GOTO 350  ! GO TO THE PRIMAL ALGORITHM
      ENDIF
C
C STEP 0 IN [2]
C DETERMINE THE FIRST THREE MINIMA OF EACH COLUMN AND DEFINE AN OPTIMAL
C SOLUTION OF CARDINALITY NASS .LE. K AND COST Z
C
      DO J = 1, N
         C(J) = J
         DO IX = B_FIRST(J), B_FIRST(J+1)-1
            IA = B_COST(IX)
            IF ( IA .LT. TMINC(J) ) THEN
               IF ( IA .LT. SMINC(J) ) THEN
                  IF ( IA .LT. MINC(J) ) THEN
                     TMINC(J)  = SMINC(J)
                     PTMINC(J) = PSMINC(J)
                     SMINC(J)  = MINC(J)
                     PSMINC(J) = PMINC(J)
                     MINC(J)   = IA
                     PMINC(J)  = B_TAIL(IX) 
                  ELSE
                     TMINC(J)  = SMINC(J)
                     PTMINC(J) = PSMINC(J)
                     SMINC(J)  = IA
                     PSMINC(J) = B_TAIL(IX)
                 ENDIF
               ELSE
                 TMINC(J)  = IA
                 PTMINC(J) = B_TAIL(IX)
               ENDIF            
            ENDIF
         ENDDO
      ENDDO
      CALL SORTI(N,MINC,C,ORDN)
C
C EXECUTE THE SAME STEPS FOR TRANSPOSED MATRIX
C COMPUTE THE FIRST THREE MINIMA OF EACH ROW
C
      DO I=1,M
         R(I) = I
         DO IX=F_FIRST(I),F_FIRST(I+1)-1
            IA = F_COST(IX)
            IF (IA .LT. TMINR(I)) THEN
               IF ( IA .LT. SMINR(I) ) THEN
                  IF ( IA .LT. MINR(I) ) THEN
                     TMINR(I)  = SMINR(I)
                     PTMINR(I) = PSMINR(I)
                     SMINR(I)  = MINR(I)
                     PSMINR(I) = PMINR(I)
                     MINR(I)   = IA
                     PMINR(I)  = F_HEAD(IX)
                  ELSE
                     TMINR(I)  = SMINR(I)
                     PTMINR(I) = PSMINR(I)
                     SMINR(I)   = IA
                     PSMINR(I)  = F_HEAD(IX)
                  ENDIF
               ELSE
                  TMINR(I)  = IA
                  PTMINR(I) = F_HEAD(IX)
               ENDIF
            ENDIF
         ENDDO
      ENDDO
      CALL SORTI(M,MINR,R,ORDM)
C
C COMPUTE A TRIVIAL LOWER BOUND LB AS THE SMALLEST VALUE BETWEEN THE SUMS 
C OF THE SMALLEST K COLUMN AND K ROW MINIMA 
C
      SUMC = MINC(C(1))
      DO JJ = 2, K-1
         SUMC = SUMC + MINC(C(JJ))
      ENDDO 
      LB_DELTA = SUMC
      SUMC     = SUMC + MINC(C(K))
      LB       = SUMC
C 
      SUMR = MINR(R(1))
      DO II=2,K-1
         SUMR = SUMR + MINR(R(II))
      ENDDO 
      IF (SUMR .GT. LB_DELTA) LB_DELTA = SUMR
      SUMR = SUMR + MINR(R(K))
      IF ( LB .LT. SUMR ) LB = SUMR
C
C STEP 1 IN [2]
C COMPUTE A PARTIAL OPTIMAL SOLUTION  OF VALUE Z AND CARDINALITY NASS (.LE. K)
C
      JMAX = 0
      UBMU = INF
      DO 90 JJ=1,N
         J = C(JJ)
         IF (MINC(J) .GT. UBMU) GOTO 100
         IF (ROW(J) .GT. 0) GOTO 90
         I = PMINC(J)
         IF ( COL(I) .GT. 0 ) GO TO 10
            COL(I) = J
            ROW(J) = I
            MU = MINC(J)
            JMAX = JJ
            GOTO 80
C
10          IF (JJ .EQ. N) GOTO 100
            L = COL(I)
            IQ = C(JJ+1)
            DO IX=B_FIRST(L),B_FIRST(L+1)-1
               IF (B_TAIL(IX) .EQ. I) THEN
                   IA = B_COST(IX)
                   GOTO 20
               ENDIF
            ENDDO
20          IF (MINC(L) .LT. IA)GOTO 50
C
C CASE NUMBER 1
C
            IF (SMINC(J) .GE. MINC(IQ) .AND. COL(PMINC(IQ)) .EQ. 0 )
     +          GOTO 30
            IF (SMINC(J) .LE. MINC(IQ) .AND. COL(PSMINC(J)) .EQ. 0 ) 
     +          GOTO 40
            GOTO 100
C
C CASE NUMBER 1.A
C
30          MIN = SMINC(L)
            IF (MIN .GT. SMINC(J)) MIN = SMINC(J)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (MINC(IQ) .GT. MIN) GOTO 100
            ROW(IQ) = PMINC(IQ)
            COL(PMINC(IQ)) = IQ
            UBMU = MIN 
            V(J) = -1
            V(L) = -1
            MU = MINC(IQ)
            JMAX=JJ+1
            GOTO 80
C
C CASE NUMBER 1.B
C
40          MIN = SMINC(L)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (SMINC(J) .GT. MIN) GOTO 100
            ROW(J) = PSMINC(J)
            COL(PSMINC(J)) = J
            UBMU = MIN
            IF (UBMU .GT. TMINC(J)) UBMU = TMINC(J)
            V(J) = -1
            V(L) = -1         
            MU = SMINC(J)
            JMAX=JJ
            GOTO 80
C
C CASE NUMBER 2 (COLUMN L WAS  ASSIGNED TO THE SECOND MINIMUM BY CASE 1)
C
50          CONTINUE    
            IF (SMINC(J) .GE. MINC(IQ) .AND. COL(PMINC(IQ)) .EQ. 0 )
     +          GOTO 60
            IF (SMINC(J) .LE. MINC(IQ) .AND. COL(PSMINC(J)) .EQ. 0 ) 
     +          GOTO 70
            GOTO 100
C
C CASE NUMBER 2.A
C
60          MIN = SMINC(J)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (MINC(IQ) .GT. MIN) GOTO 100
            ROW(IQ) = PMINC(IQ)
            COL(PMINC(IQ)) = IQ
            UBMU = MIN 
            V(J) = -1
            IF ( V(L) .NE. -1) STOP ' V(L) .NE. -1 (CASO A)'
            MU = MINC(IQ)
            JMAX=JJ+1
            GOTO 80
C
C CASE NUMBER 2.B
C
70          IF (SMINC(J) .GT. UBMU) GOTO 100
            ROW(J) = PSMINC(J)
            COL(PSMINC(J)) = J
            IF (UBMU .GT. TMINC(J)) UBMU = TMINC(J)
            V(J) = -1
            IF (V(L) .NE. -1) STOP ' V(L) .NE. -1 (CASO B)'
            MU = SMINC(J)
            JMAX=JJ
C
80          NASS = NASS + 1
            IF ( NASS .EQ. K ) GOTO 100
90    CONTINUE
C
100   Z = 0
      DO JJ=1,JMAX
         J = C(JJ)
         I = ROW(J)
         IF (I .GT. 0) THEN 
            DO IX=B_FIRST(J),B_FIRST(J+1)-1
               IF (B_TAIL(IX) .EQ. I) THEN
                  IA = B_COST(IX)
                  GOTO 110
               ENDIF
            ENDDO 
110         IF (V(J) .LT. 0) THEN
               V(J) = 0
               U(I) = MU - IA
               Z = Z + IA
            ELSE
               V(J) = MU - IA
               U(I) = 0
               Z = Z + IA
            ENDIF
         ELSE
            V(J) = 0
         ENDIF
      ENDDO
C
      NASS1 = NASS
      IF ( NASS .EQ. K ) GO TO 390
C
C DETERMINE A FIRST GREEDY SOLUTION
C      
      HNASS = NASS
      HZ = Z
      CALL GREEDY(M,N,B_COST,B_TAIL,B_FIRST,INF,K,ROW,COL,R,
     +     PMINR,PSMINR,C,PMINC,PSMINC,HZ,HNASS,1,JMAX,PTMINR,PTMINC)
      IF ( HNASS .LT. K ) HZ = INF
      IF ( HZ .EQ. LB ) GOTO 230      
C 
C STEP 2 IN [2]
C COMPUTE A PARTIAL OPTIMAL SOLUTION OF VALUE Z2 AND CARDINALITY NASS2 (.LE. K)
C
      MU2   = 0
      NASS2 = 0
      Z2    = 0
      IMAX = 0
      UBMU = INF
      DO 200 II=1,M
         I = R(II)
         IF (MINR(I) .GT. UBMU) GOTO 210
         IF (COL2(I) .GT. 0) GOTO 200
         J = PMINR(I)
         IF ( ROW2(J) .GT. 0 ) GO TO 120
            ROW2(J) = I
            COL2(I) = J
            MU2 = MINR(I)
            IMAX = II
            GOTO 190
120         IF (II .EQ. M) GOTO 210
            L = ROW2(J)
            DO IX=B_FIRST(J),B_FIRST(J+1)-1
               IF (B_TAIL(IX) .EQ. L) THEN
                  IA = B_COST(IX)
                  GOTO 130
               ENDIF
            ENDDO
130         IF (MINR(L) .LT. IA) GOTO 160
            IQ = R(II+1)
C
C CASE NUMBER 1
C
            IF (SMINR(I) .GE. MINR(IQ) .AND. ROW2(PMINR(IQ)) .EQ. 0 )
     +          GOTO 140
            IF (SMINR(I) .LE. MINR(IQ) .AND. ROW2(PSMINR(I)) .EQ. 0 ) 
     +          GOTO 150
            GOTO 210
C
C CASE NUMBER 1.A
C
140         MIN = SMINR(L)
            IF (MIN .GT. SMINR(I)) MIN = SMINR(I)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (MINR(IQ) .GT. MIN) GOTO 210
            COL2(IQ) = PMINR(IQ)
            ROW2(PMINR(IQ)) = IQ
            UBMU = MIN 
            U2(I) = -1
            U2(L) = -1
            MU2 = MINR(IQ)
            IMAX=II+1
            GOTO 190
C
C CASE NUMBER 1.B
C
150         MIN = SMINR(L)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (SMINR(I) .GT. MIN) GOTO 210
            COL2(I) = PSMINR(I)
            ROW2(PSMINR(I)) = I
            UBMU = MIN
            IF (UBMU .GT. TMINR(I)) UBMU = TMINR(I)
            U2(I) = -1
            U2(L) = -1         
            MU2 = SMINR(I)
            IMAX=II
            GOTO 190
C
C CASE NUMBER 2 (ROW L WAS ASSIGNED TO THE SECOND MINIMUM BY CASE 1)
C
160         CONTINUE    
            IF (SMINR(I) .GE. MINR(IQ) .AND. ROW2(PMINR(IQ)) .EQ. 0 )
     +          GOTO 170
            IF (SMINR(I) .LE. MINR(IQ) .AND. ROW2(PSMINR(I)) .EQ. 0 ) 
     +          GOTO 180
            GOTO 210
C
C CASE NUMBER 2.A
C
170         MIN = SMINR(I)
            IF (MIN .GT. UBMU) MIN = UBMU
            IF (MINR(IQ) .GT. MIN) GOTO 210
            COL2(IQ) = PMINR(IQ)
            ROW2(PMINR(IQ)) = IQ
            UBMU = MIN 
            U2(I) = -1
            MU2 = MINR(IQ)           
            IMAX=II+1
            GOTO 190
C
C CASE NUMBER 2.B
C
180         IF (SMINR(I) .GT. UBMU) GOTO 210
            COL2(I) = PSMINR(I)
            ROW2(PSMINR(I)) = I
            IF (UBMU .GT. TMINR(I)) UBMU = TMINR(I)
            U2(I) = -1
            MU2 = SMINR(I)
            IMAX=II
190         NASS2 = NASS2 + 1
            IF ( NASS2 .EQ. K ) GO TO 210
200   CONTINUE
C
210   Z2 = 0  
      DO II=1,IMAX
         I = R(II)
         J = COL2(I)
         IF (J .GT. 0) THEN 
            DO IX=B_FIRST(J),B_FIRST(J+1)-1
               IF (B_TAIL(IX) .EQ. I) THEN
                  IA = B_COST(IX)
                  GOTO 220
               ENDIF
            ENDDO
220         IF (U2(I) .LT. 0) THEN
               U2(I) = 0
               V2(J) = MU2 - IA
               Z2 = Z2 + IA 
            ELSE
               U2(I) = MU2 - IA
               V2(J) = 0
               Z2 = Z2 + IA
            ENDIF
         ELSE
            U2(I) = 0
         ENDIF
      ENDDO
C
      IF ( NASS2 .EQ. K ) GO TO 380
C
C   DETERMINE A SECOND GREEDY SOLUTION
C
      HNASS2 = NASS2
      HZ2 = Z2
      CALL GREEDY(M,N,B_COST,B_TAIL,B_FIRST,INF,K,ROW2,COL2,R,
     +     PMINR,PSMINR,C,PMINC,PSMINC,HZ2,HNASS2,2,IMAX,PTMINR,PTMINC)
      IF ( HNASS2 .LT. K ) HZ2 = INF
C
C STEP 3 IN [2]
C HERE WE HAVE COMPUTED:
C a) A TRIVIAL LOWER BOUND OF VALUE LB 
C b) TWO HEURISTIC SOLUTIONS OF VALUE HZ AND HZ2, RESPECTIVELY
C c) TWO PARTIAL OPTIMAL SOLUTIONS WITH NASS AND NASS2 ELEMENTS, RESPECTIVELY
C    (MAX(NASS,NASS2) .LT. K)
C d) TWO PARTIAL OPTIMAL SOLUTION OF CARDINALITY NASS AND NASS2
C   
C SELECT THE BEST HEURISTIC SOLUTION
C
230   IF ( HZ2 .LT. HZ .OR. HNASS2 .GT. HNASS) THEN
         DO J=1,N
            IF ( ROW2(J) .NE. 0 ) THEN
               HROW(J) = IABS(ROW2(J))
            ELSE
               HROW(J) = 0
            ENDIF   
         ENDDO
         DO I=1,M
            IF ( COL2(I) .NE. 0 ) THEN
               HCOL(I) = IABS(COL2(I))
            ELSE
               HCOL(I) = 0
            ENDIF   
         ENDDO
         HZ    = HZ2
         HNASS = HNASS2
      ELSE
         DO J=1,N
            IF ( ROW(J) .NE. 0 ) THEN
               HROW(J) = IABS(ROW(J))
            ELSE
               HROW(J) = 0
            ENDIF   
         ENDDO
         DO I=1,M
            IF ( COL(I) .NE. 0 ) THEN
               HCOL(I) = IABS(COL(I))
            ELSE
               HCOL(I) = 0
            ENDIF   
         ENDDO
      ENDIF
      IF ( HZ .EQ. LB ) GOTO 400
C
C SELECT THE BEST PARTIAL OPTIMAL SOLUTION
C
      IF ( NASS2 .GT. NASS ) THEN
         NASS = NASS2
         DO J=1,N
            V(J) = V2(J)
            IF ( ROW2(J) .GT. 0 ) THEN
               ROW(J) = ROW2(J)
            ELSE
               ROW(J) = 0
            ENDIF   
         ENDDO
         DO I=1,M
            U(I) = U2(I)
            IF ( COL2(I) .GT. 0 ) THEN
               COL(I) = COL2(I)
            ELSE
               COL(I) = 0
            ENDIF   
         ENDDO
         Z = Z2
         MU = MU2
      ELSE
         DO J=1,N
            IF ( ROW(J) .LT. 0 ) ROW(J) = 0
         ENDDO
         DO I=1,M
            IF ( COL(I) .LT. 0 ) COL(I) = 0
         ENDDO
      ENDIF
c
      IF ( HZ .EQ. INF ) THEN
C
C COMPLETE THE BEST HEURISTIC SOLUTION FOUND IN A GREEDY WAY
C
         HN = HNASS
         DO 310 J=1,N
            IF ( HROW(J) .EQ. 0 ) THEN
               DO IX=B_FIRST(J),B_FIRST(J+1)-1
                  I = B_TAIL(IX)
                  IF ( HCOL(I) .EQ. 0 ) THEN
                     HN = HN + 1
                     HCOL(I) = J
                     HROW(J) = I
                     IF ( HN .LT. K ) GOTO 310
                     GOTO 340
                  ENDIF
               ENDDO
            ENDIF
 310     CONTINUE
         IF ( HN .NE. K ) THEN
             DO 320 J=1,N
                IF ( HROW(J) .EQ. 0 ) THEN
                   I = 0
 330               I = I + 1
                   IF ( HCOL(I) .EQ. 0 ) THEN
                      HN = HN + 1
                      HCOL(I) = J
                      HROW(J) = I
                      IF (HN .LT. K) GOTO 320
                      GOTO 340
                   ENDIF
                   IF ( I .LT. M ) GOTO 330
                ENDIF
 320         CONTINUE
         ENDIF
      ENDIF     
C
C STEP 4 IN [2]
C DETERMINE THE "FIXED" ROWS AND COLUMNS OF A PARTIAL OPTIMAL SOLUTION
C AND STORE THEM INTO SETFROW(1,...,NSETFROW), SETFCOL(1,...,NSETFCOL) 
C ( PROPERTY 4 IN [1] )
C
 340  NSETFROW = 0
      NSETFCOL = 0
      DO I=1,M
         FROW(I) = .FALSE.
      ENDDO
      DO J=1,N
         IF (ROW(J) .EQ. 0) THEN
             FCOL(J) = .FALSE.
         ELSE
             FCOL(J)           = .TRUE.
             NSETFCOL          = NSETFCOL + 1
             SETFCOL(NSETFCOL) = J
             I                 = ROW(J)
             FROW(I)           = .TRUE.
             NSETFROW          = NSETFROW + 1
             SETFROW(NSETFROW) = I
         ENDIF
      ENDDO
C
C "FIXED" ROWS (PROPERTY 5 IN [1] )
C
      IQ = NSETFCOL
      DO JJ=1,K
         J = C(JJ)      
         I = PMINC(J)
         IF (.NOT. FCOL(J)) IQ = IQ + 1
         IF (ROW(J) .EQ. 0) THEN
             IF (.NOT.FROW(I)) THEN
                 FROW(I) = .TRUE.
                 NSETFROW = NSETFROW + 1
                 SETFROW(NSETFROW) = I
             ENDIF
         ENDIF
         IF (IQ .EQ. K) GOTO 240
      ENDDO  
240   CONTINUE
C
C ( PROPERTY 6 IN [1] )
C
      DO IQ=1,NSETFROW
         I = SETFROW(IQ)
         IF (COL(I) .EQ. 0) THEN
             J = PMINR(I)
             IF (.NOT.FCOL(J)) THEN
                 FCOL(J) = .TRUE.
                 NSETFCOL = NSETFCOL + 1
                 SETFCOL(NSETFCOL) = J
             ENDIF
         ENDIF
      ENDDO
      SET = 1
      IF (NSETFROW .EQ. K .AND. NSETFCOL .EQ. K) GOTO 370
C
C "FIXED" COLUMNS (PROPERTY 8  IN [1] )
C
      IQ = NSETFROW
      DO II=1,K
         I = R(II)      
         J = PMINR(I)
         IF (.NOT. FROW(I)) IQ = IQ + 1
         IF (COL(I) .EQ. 0) THEN
             IF (.NOT.FCOL(J)) THEN
                 FCOL(J) = .TRUE.
                 NSETFCOL = NSETFCOL + 1
                 SETFCOL(NSETFCOL) = J
             ENDIF
         ENDIF
         IF (IQ.EQ.K) GOTO 250
      ENDDO
250   CONTINUE
C
C ( PROPERTY 8  IN [1] )
C
      DO IQ = 1,NSETFCOL
         J = SETFCOL(IQ)
         IF (ROW(J) .EQ. 0) THEN
             I = PMINC(J)
             IF (.NOT.FROW(I)) THEN
                 FROW(I) = .TRUE.
                 NSETFROW = NSETFROW + 1
                 SETFROW(NSETFROW) = I
             ENDIF
         ENDIF
      ENDDO
C
      SET = 1
      IF (NSETFROW .EQ. K .AND. NSETFCOL .EQ. K) GOTO 370
C
C COMPUTE A NEW LOWER BOUND (SEE [1], SECTION 4.2)
C
      CALL LOWERB(M,N,K,FROW,FCOL,NSETFROW,NSETFCOL,
     + MINR,PMINR,MINC,PMINC,SMINR,SMINC,R,C,INF,HZ,LB)
      IF (LB .EQ. HZ) GOTO 400
C
C DETERMINE THE "FIXED" ROWS AND COLUMNS USING THE TRANSPOSE MATRIX
C AND STORE THEM INTO SETFROW2(1,...,NSETFROW2), SETFCOL2(1,...,NSETFCOL2) 
C
C ( PROPERTY 4 IN [1] )
C
      NSETFROW2 = 0
      NSETFCOL2 = 0
      DO I=1,M
         FROW2(I) = .FALSE.
      ENDDO
      DO J=1,N
         IF (ROW(J) .EQ. 0) THEN
             FCOL2(J)=.FALSE.
         ELSE
             FCOL2(J)=.TRUE.
             NSETFCOL2 = NSETFCOL2 + 1
             SETFCOL2(NSETFCOL2) = J
             I = ROW(J)
             FROW2(I)=.TRUE.
             NSETFROW2 = NSETFROW2 + 1
             SETFROW2(NSETFROW2) = I
         ENDIF
      ENDDO
C
C "FIXED" COLUMNS (PROPERTY 5  IN [1] )
C
      IQ = NSETFROW2
      DO II=1,K
         I = R(II)
         J = PMINR(I)
         IF (.NOT. FROW2(I)) IQ = IQ + 1
         IF (COL(I) .EQ. 0) THEN
             IF (.NOT.FCOL2(J)) THEN
                 FCOL2(J) = .TRUE.
                 NSETFCOL2 = NSETFCOL2 + 1
                 SETFCOL2(NSETFCOL2) = J
             ENDIF
         ENDIF
         IF (IQ .EQ. K) GOTO 260
      ENDDO  
260   CONTINUE
C
C ( PROPERTY 6 IN [1] )
C
      DO IQ=1,NSETFCOL2
         J = SETFCOL2(IQ)
         IF (ROW(J) .EQ. 0) THEN
             I = PMINC(J)
             IF (.NOT.FROW2(I)) THEN
                 FROW2(I) = .TRUE.
                 NSETFROW2 = NSETFROW2 + 1
                 SETFROW2(NSETFROW2) = I
              ENDIF
         ENDIF
      ENDDO
      SET = 2
      IF (NSETFROW2 .EQ. K .AND. NSETFCOL2 .EQ. K) GOTO 370
C
C "FIXED" ROWS (PROPERTY 7  IN [1] )
C
      IQ = NSETFCOL2
      DO JJ=1,K
         J = C(JJ)
         I = PMINC(J)
         IF (.NOT. FCOL2(J)) IQ = IQ + 1
         IF (ROW(J) .EQ. 0) THEN
             IF (.NOT.FROW2(I)) THEN
                 FROW2(I) = .TRUE.
                 NSETFROW2 = NSETFROW2 + 1
                 SETFROW2(NSETFROW2) = I
             ENDIF
         ENDIF
         IF (IQ.EQ.K) GOTO  270
      ENDDO
270   CONTINUE
C
C (PROPERTY 8 IN [1] )
C
      DO IQ = 1,NSETFROW2
         I = SETFROW2(IQ)
         IF (COL(I) .EQ. 0) THEN         
             J = PMINR(I)
             IF (.NOT.FCOL2(J)) THEN
                 FCOL2(J) = .TRUE.
                 NSETFCOL2 = NSETFCOL2 + 1
                 SETFCOL2(NSETFCOL2) = J
             ENDIF
         ENDIF
      ENDDO
C
      SET = 2
      IF (NSETFROW2 .EQ. K .AND. NSETFCOL2 .EQ. K) GOTO 370
C 
C COMPUTE THE LOWER BOUND ON THE TRANSPOSED MATRIX
C
      CALL LOWERB(M,N,K,FROW2,FCOL2,NSETFROW2,NSETFCOL2,
     + MINR,PMINR,MINC,PMINC,SMINR,SMINC,R,C,INF,HZ,LB2)

C
C CHOOSE THE BEST LOWER BOUND
C
      IF (LB2 .GT. LB) LB = LB2
      IF (LB .EQ. HZ) GOTO 400
C      
      DELTA = HZ - LB_DELTA
      DZ    = HZ - Z
C
C PRIMAL ALGORITHM
C BUILD THE LOCAL MATRIX (CALLED W IN [2])
C
350   CALL DEFW(M,N,F_COST,F_HEAD,F_FIRST,B_COST,B_TAIL,B_FIRST,
     +    HROW,MU,U,V,DELTA,DZ,
     +    R,C,INVR,INVC,F_COST2,F_HEAD2,F_FIRST2,F_LAST2,
     +    B_COST2,B_TAIL2,B_FIRST2,
     +    B_LAST2,RID,NUMR,NUMC)
C
C SOLVE THE AP ON THE REDUCED MATRIX WITH  A MODIFIED VERSION OF ALGORITHM JV 
C (JONKER & VOLGEANT, COMPUTING 38, 1987)
C NAF  = NUMBER OF DUMMY ARCS INTRODUCED BY SPJV_MOD
C
      CALL SPJV_MOD(K,B_COST2,B_TAIL2,B_FIRST2,B_LAST2,
     +      ROW2,COL2,U2,V2,Z,Y1,Y2,Y3,Y4,OK,INF,BIGM,NAF)
C
C DEFINE THE CORRESPONDING KAP'S DUAL SOLUTION
C
      ALFA = 0
      BETA = 0
      DO I = 1,K
         IF (ALFA .LT. U2(I)) ALFA = U2(I)
         IF (BETA .LT. V2(I)) BETA = V2(I)
      ENDDO
      MU = ALFA + BETA
      DO I=1,K
         U2(I) = ALFA - U2(I)
         V2(I) = BETA - V2(I)
      ENDDO
C
C MAXROW = MAXIMUM ROW CONSIDERED AND INSERTED IN THE SPARSE MATRIX
C MAXCOL = MAXIMUM COL CONSIDERED AND INSERTED IN THE SPARSE MATRIX 
C
      MAXROW = K
      MAXCOL = K
      DO II=1,N-K
         IF (II .GT.M-K) GOTO 360
         I = R(K + II)
         CALL RHEAP(F_COST,F_HEAD,F_FIRST,F_COST2,F_HEAD2,F_FIRST2,
     +           F_LAST2,MAXROW,B_COST2,B_TAIL2,B_FIRST2,B_LAST2,MAXCOL,
     +           ROW2,COL2,MU,U2,V2,Z,INVC,R,INVR,NUMR,
     +           BIGM,RID,NAF,DELTA,INF)
360      J = C(K + II)
         CALL CHEAP(B_COST,B_TAIL,B_FIRST,F_COST2,F_HEAD2,F_FIRST2,
     +           F_LAST2,MAXROW,B_COST2,B_TAIL2,B_FIRST2,B_LAST2,MAXCOL,
     +           ROW2,COL2,MU,U2,V2,Z,INVR,C,INVC,NUMC,
     +           BIGM,RID,NAF,DELTA,INF)
      ENDDO
C
      IF (NAF .GT. 0) RETURN
C     
C RID = NUMBER OF ELEMENTS REMOVED FROM THE ORIGINAL MATRIX.
C NB: IF RID .GT. 0 THE DUAL SOLUTION COULD BE NOT OPTIMAL FOR THE ORIGINAL MATRIX
C
      OPT = 1
      DO II=1,M
         I    = R(II)
         U(I) = U2(II)
         JJ   = COL2(II)
         IF (JJ .GT. 0) THEN 
            J = C(JJ)
         ELSE
            J = 0
         ENDIF
         COL(I) = J
      ENDDO
      DO JJ=1,N
         J    = C(JJ)
         V(J) = V2(JJ)
         II   = ROW2(JJ)
         IF (II .GT. 0) THEN 
             I = R(II)
         ELSE
             I = 0
         ENDIF
         ROW(J) = I
      ENDDO
      RETURN
C
C SOLVE A SQUARE ASSIGNMENT WITH THE K IMPOSED ROWS AND COLUMNS
C NB: THE DUAL SOLUTION COULD BE NOT OPTIMAL FOR THE ORIGINAL MATRIX
C
 370  OPT = 1
      DO J=1,N
         V(J) = 0
         ROW(J) = 0
         HROW(J) = 0
      ENDDO
      DO I=1,M
         U(I) = 0
         COL(I) = 0
         HCOL(I) = 0
      ENDDO
      IF (SET .EQ. 1) THEN
         CALL SPJV_PUNT(K,F_COST,F_HEAD,B_COST,B_TAIL,F_FIRST,B_FIRST,
     +        R,C,U,V,Z,MINC,MINR,PMINC,PMINR,SMINR,
     +        INF,SETFROW,HCOL,SETFCOL,HROW)
      ELSE
         CALL SPJV_PUNT(K,F_COST,F_HEAD,B_COST,B_TAIL,F_FIRST,B_FIRST,
     +        R,C,U,V,Z,MINC,MINR,PMINC,PMINR,SMINR,
     +        INF,SETFROW2,HCOL,SETFCOL2,HROW)
      ENDIF
C
      DO II =1,K
         I  = SETFROW(II)
         JJ = C(II)
         J  = SETFCOL(JJ)
         COL(I) = J
         ROW(J) = I
      ENDDO
      RETURN
C
C STORE THE SOLUTION OF VALUE Z2 OBTAINED AS PARTIAL OPTIMAL SOLUTION 
C DETERMINED USING THE TRANSPOSED MATRIX
C
380   OPT = 1
      Z   = Z2
      MU  = MU2
      NASS1 = NASS2
      DO I=1,M
         COL(I) = COL2(I)
         U(I)   = U2(I)
      ENDDO
      DO J=1,N
         V(J)   = V2(J)
         ROW(J) = ROW2(J)         
      ENDDO
      RETURN
C
C RETURN THE FIRST PARTIAL OPTIMAL SOLUTION
C
390   OPT = 1
      RETURN
C
C SUBROUTINE GREEDY HAVE FOUND THE OPTIMAL SOLUTION
C NB: THE DUAL SOLUTION COULD BE NOT OPTIMAL FOR THE ORIGINAL MATRIX
C
400   OPT = 1
      Z   = HZ
      DO I=1,M
         COL(I) = HCOL(I)
      ENDDO
      DO J=1,N
         ROW(J) = HROW(J)
      ENDDO     
      RETURN
      END
C===============================================================================
      SUBROUTINE DEFW(M,N,F_COST,F_HEAD,F_FIRST,B_COST,B_TAIL,
     +    B_FIRST,HROW,MU,U,V,DELTA,DZ,
     +    R,C,INVR,INVC,F_COST2,F_HEAD2,F_FIRST2,F_LAST2,
     +    B_COST2,B_TAIL2,B_FIRST2,
     +    B_LAST2,RID,NUMR,NUMC)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
C
C THIS SUBROUTINE DEFINES THE LOCAL MATRIX (CALLED W IN [2]) 
C
C INPUT**************************************************************************
C M         = NUMBER OF ROWS
C N         = NUMBER OF COLUMNS
C F_COST,F_HEAD,F_FIRST = FORWARD STAR OF THE ORIGINAL MATRIX
C B_COST,B_TAIL,B_FIRST = BACKWARD STAR OF THE ORIGINAL MATRIX
C HROW      = HEURISTIC SOLUTION
C MU,U,V    = DUAL VARIABLES OF THE BEST PARTIAL OPTIMAL SOLUTION 
C DELTA,DZ  = UPPER BOUNDS TO THE VALUE OF AN ELEMENT TO BE INSERTED IN SOLUTION
C
C OUTPUT*************************************************************************
C R(1,...,K)    = ROWS TO BE INCLUDED IN THE REDUCED MATRIX
C R(K+1,...,M)  = NOT ASSIGNED ROWS
C C(1,...,K)    = COLUMNS TO BE INCLUDED IN THE REDUCED MATRIX
C C(K+1,...,N)  = NOT ASSIGNED COLUMNS
C INVR(1,...,M) = ROWS' INVERTED INDEX: ROWS( INVR(H) ) = H 
C INVC(1,...,N) = COLUMNS' INVERTED INDEX
C F_COST2,F_HEAD2,F_FIRST2,F_LAST2 = FORWARD STAR OF THE REDUCED MATRIX
C B_COST2,B_TAIL2,B_FIRST2,B_LAST2 = BACKWARD STAR OF THE REDUCED MATRIX
C RID           = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX WHICH HAVE 
C                 NOT BEEN INSERTED
C NUMR,NUMC     = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX IN THE K ROWS (COLUMNS)
C                 OF THE REDUCED MATRIX
C
      INTEGER*4 M,N,F_COST(ORDS),B_COST(ORDS),
     + F_FIRST(ORDM+1),B_FIRST(ORDN+1),
     + B_COST2(ORDS),F_COST2(ORDS),B_FIRST2(N),F_FIRST2(M),U(ORDM),
     + V(ORDN),MU,B_LAST2(N),F_LAST2(M),HROW(N),R(M),C(N),
     + INVR(M),INVC(N),NUMR,NUMC,DELTA,DZ
      INTEGER*2 F_HEAD(ORDS),B_TAIL(ORDS),B_TAIL2(ORDS),F_HEAD2(ORDS)
      LOGICAL F(ORDM)
      INTEGER*4 I,IB,IS,II,IX,J,JJ,MAX,NN,RID,NEL,IIS,EL,IEL
      INTEGER*4 INDC(ORDN),INDR(ORDM),JMU,IMU
      EQUIVALENCE (INDC(1),INDR(1))
C
      DO I=1,M
         F_FIRST2(I) = 0
         F_LAST2(I) = 0
      ENDDO
      DO J=1,N
         B_FIRST2(J) = 0
         B_LAST2(J) = 0
      ENDDO
      RID = 0
C
      DO I=1,M
         INVR(I) = -1 
      ENDDO
      DO J=1,N
         INVC(J) = -1
      ENDDO
      MAX = 0
      NN  = N + 1
      DO 40 J = 1,N
         I = HROW(J)
         IF (I .EQ. 0) THEN
            NN      = NN - 1
            C(NN)   = J
            INVC(J) = NN
         ELSE
            IF (I .LT. 0) STOP 
            MAX     = MAX + 1
            R(MAX)  = I
            C(MAX)  = J
            INVR(I) = MAX
            INVC(J) = MAX
         ENDIF
40    CONTINUE
      NN = M+ 1
      DO 50 I=1,M
         IF (INVR(I) .GT. 0) GOTO 50
         NN       = NN - 1
         R(NN) = I
         INVR(I)  = NN
50    CONTINUE
C
C COPY THE MATRIX (COLUMN STRUCTURE)
C
      DO JJ=1,MAX
         F(JJ) = .FALSE.
      ENDDO
      NEL = 0
      NUMC = 0
      DO JJ =1,MAX
         J    = C(JJ)
         IMU  = -MU + V(J)
         IB   = NUMC 
         NEL  = B_FIRST(J+1) - B_FIRST(J)
         NUMC = NUMC + NEL
         B_FIRST2(JJ) = IB + 1
         DO IX=B_FIRST(J),B_FIRST(J+1)-1
            I = B_TAIL(IX)
            IF (INVR(I) .LE. MAX) THEN
               IS  = B_COST(IX)
               IIS = INVR(I)
               IF ( IS.LE.DELTA .AND. IS+IMU+U(I).LE.DZ ) THEN 
                  IB    = IB + 1
                  B_LAST2(JJ) = IB
                  B_TAIL2(IB) = IIS
                  B_COST2(IB) = IS
               ELSE
                  RID = RID + 1
                  EL = IS
                  IEL = IIS
               ENDIF
            ENDIF
         ENDDO
C
C TO AVOID EMPTY COLUMNS STORE ONE OF THE DISCHARGED ELEMENTS
C
         IF (B_LAST2(JJ) .EQ. 0) THEN
            F(JJ) = .TRUE.
            IB = IB + 1
            B_LAST2(JJ) = IB
            B_TAIL2(IB) = IEL
            B_COST2(IB) = EL
            RID = RID - 1
         ENDIF
      ENDDO
C
C BUILD THE FORWARD STAR
C
      NEL = 0
      NUMR = 0
      DO II=1,MAX
         I    = R(II)
         JMU  = -MU +U(I)
         IB   = NUMR
         NEL  = F_FIRST(I+1) - F_FIRST(I)
         NUMR = NUMR + NEL
         F_FIRST2(II) = IB+1
         DO IX=F_FIRST(I),F_FIRST(I+1)-1
            J = F_HEAD(IX)
            IF (INVC(J) .LE. MAX) THEN
               IS = F_COST(IX)
               IF (IS.LE.DELTA .AND. IS+JMU+V(J).LE.DZ) THEN
                  IB = IB + 1
                  F_LAST2(II) = IB
                  F_HEAD2(IB) = INVC(J)
                  F_COST2(IB) = IS
               ENDIF
            ENDIF
         ENDDO
      ENDDO
C
C ADD THE ELEMENTS OF THE BACKWARD STAR THAT HAVE BEEN FIRST DISCHARGED AND 
C THEN STORED TO AVOID EMPTY COLUMNS
C
      DO JJ=1,MAX
         IF (F(JJ)) THEN
            IX = B_FIRST2(JJ)
            II = B_TAIL2(IX)
            IB = F_LAST2(II) + 1
            IF (IB .EQ. 1) IB = F_FIRST2(II)
            F_LAST2(II) = IB
            F_HEAD2(IB) = JJ
            F_COST2(IB) = B_COST2(IX)
         ENDIF
      ENDDO
      RETURN
      END
C=================================================================================
      SUBROUTINE RHEAP(F_COST,F_HEAD,F_FIRST,F_COST2,F_HEAD2,
     +           F_FIRST2,F_LAST2,MAXROW,B_COST2,B_TAIL2,B_FIRST2,
     +           B_LAST2,MAXCOL,ROW,COL,MU,U,V,Z,INVC,R,INVR,
     +           NUMR,BIGM,RID,NAF,DELTA,INF)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)      
C
C THIS SUBROUTINE ADDS THE (MAXROW+1)-TH ROW TO THE SPARSE MATRIX AND
C REOPTIMIZE THE CURRENT SOLUTION
C
C INPUT****************************************************************************
C F_COST,F_HEAD,F_FIRST = FORWARD STAR OF THE ORIGINAL MATRIX
C F_COST2,F_HEAD2,F_FIRST2,F_LAST2 = FORWARD STAR OF THE REDUCED MATRIX
C MAXROW   = NUMBER OF ROWS IN THE REDUCED MATRIX
C B_COST2,B_TAIL2,B_FIRST2,B_LAST2 = BACKWARD STAR OF THE REDUCED MATRIX
C MAXCOL   = NUMBER OF COLUMNS IN THE REDUCED MATRIX
C ROW,COL  = CURRENT PRIMAL SOLUTION
C MU, U, V = CURRENT DUAL VALUES
C R(IQ)    = INDEX OF THE ROW OF THE ORIGINAL MATRIX CORRESPONDING TO ROW IQ
C            OF THE NEW SPARSE MATRIX
C INVC()   = COLUMNS 'INVERTED' INDEX
C INVR()   = ROWS' INVERTED INDEX: ROWS( INVR(H) ) = H
C NUMR     = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX IN THE ROWS
C            OF THE REDUCED MATRIX
C BIGM     = VALUE OF THE DUMMY ELEMENTS
C RID      = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX WHICH HAVE NOT BEEN INSERTED
C NAF      = NUMBER OF DUMMY ELEMENTS
C DELTA    = UPPER BOUND TO THE VALUE OF AN ELEMENT TO BE INSERTED IN SOLUTION
C INF      = A VERY LARGE NUMBER
C
C OUTPUT***************************************************************************
C UPDATED VALUES OF THE REDUCED MATRIX AND OF THE CURRENT PRIMAL AND DUAL SOLUTIONS
C
      INTEGER*4 F_COST(ORDS),F_FIRST(ORDM+1),
     + F_COST2(ORDS),F_FIRST2(ORDM),F_LAST2(ORDM),
     + B_COST2(ORDS),B_FIRST2(ORDN),B_LAST2(ORDN),MU,U(ORDM),V(ORDN),
     + Z,MAXROW,MAXCOL,R(ORDM),INVC(ORDN),INVR(ORDM),ROW(ORDN),
     + COL(ORDM),INF,RID,NUMR,BIGM,NAF,DELTA
      INTEGER*4 LAB(ORDN),LC(ORDN),HP(ORDN),IPERM(ORDN),PERM(ORDN),
     +          Q(ORDN)
      INTEGER*2 F_HEAD(ORDS),F_HEAD2(ORDS),B_TAIL2(ORDS)
C
C LOCAL VARIABLES
C 
      INTEGER ALFA,CHK,LCJ,I,IA,J,JJ,J0,IQ,LABT,LT,H,MIN,VJ,II,IB,
     + IS,IC,NEL,NHP,IQ2,HP1,HP2,HP3,UP,IN,J1,L,IX
C
      MAXROW = MAXROW + 1
      DO J=1,MAXCOL
         LC(J)    = INF
         HP(J)    = 0
         Q(J)     = 0
         PERM(J)  = J
         IPERM(J) = J
      ENDDO
      NHP    = 0
      I      = R(MAXROW)
      INVR(I)= MAXROW
C
C SET CHK = 0 IF THE CURRENT SOLUTION REMAINS OPTIMAL
C         = -1 OTHERWHISE
C
      CHK  = 0
      IB   = NUMR
      NEL  = F_FIRST(I+1) - F_FIRST(I)
      NUMR = NUMR + NEL
      F_FIRST2(MAXROW) = IB+1
      DO IX=F_FIRST(I),F_FIRST(I+1)-1
         J  = F_HEAD(IX)
         JJ = INVC(J)
         IF (JJ .LE. MAXCOL) THEN
            IA = F_COST(IX)
            IF (IA.LE.DELTA) THEN
               IB = IB+1
               F_LAST2(MAXROW) = IB
               F_HEAD2(IB) = JJ
               F_COST2(IB) = IA
               IC = B_LAST2(JJ) + 1
               IF (IC .EQ. 1) IC = B_FIRST2(JJ)
               B_LAST2(JJ) = IC
               B_TAIL2(IC) = MAXROW
               B_COST2(IC) = IA
               LC(JJ)  = IA - MU + V(JJ)
               LAB(JJ) = MAXROW
               IF (IA - MU+ V(JJ) .LT. 0) THEN
                  CHK = -1
                  IF (Q(JJ) .EQ. 0) THEN 
                     NHP = NHP + 1
                     Q(JJ) = NHP
                  ENDIF
                  IQ = Q(JJ)
10                IQ2 = IQ/2
                  IF (IQ2 .LE. 0) GOTO 20
                  HP2 = HP(IQ2)
                  IF (LC(JJ) .GE. LC(HP2)) GOTO 20
                  HP(IQ) = HP2
                  Q(HP2) = IQ
                  IQ = IQ2
                  GOTO 10
20                HP(IQ) = JJ
                  Q(JJ) = IQ
               ENDIF
            ELSE
               RID = RID + 1
            ENDIF
         ENDIF
      ENDDO
      COL(MAXROW) = 0
      U(MAXROW)   = 0
      IF (CHK .EQ. 0) RETURN
C
C ALTERNATING PATH TO REOPTIMIZE THE SOLUTION (IF NECESSARY)
C
      UP = 1
      LT = INF
30    J0  = HP(1)
      MIN = LC(J0)
      IF (MIN .GE. LT) GOTO 150
      IN = IPERM(J0)
      J1 = PERM(UP)
      IPERM(J0) = UP
      IPERM(J1) = IN
      PERM(UP) = J0
      PERM(IN) = J1
      UP = UP + 1
      Q(J0)= 0
      NHP = NHP - 1
      IF (NHP .EQ. 0) GOTO 80
      HP1 = HP(NHP+1)
      LCJ = LC(HP1)
      IQ = 1
40    IQ2 = IQ*2
      IF (IQ2-NHP) 50,60,70
50    HP2 = HP(IQ2)
      HP3 = HP(IQ2+1)
      IF (LC(HP2) .LE. LC(HP3)) GOTO 60
      HP2 = HP3
      IQ2 = IQ2 + 1
60    HP2 = HP(IQ2)
      IF (LCJ .LT. LC(HP2)) GOTO 70
      HP(IQ) = HP2
      Q(HP2) = IQ
      IQ = IQ2
      GOTO 40
70    HP(IQ) = HP1
      Q(HP1) = IQ
80    I = ROW(J0)
      IF ( I .GT. 0) THEN 
         H = MIN + U(I)
         IF (H .LT. LT) THEN 
            LT   = H
            LABT = I
            IF (U(I) .EQ. 0) GOTO 150
         ENDIF  
         H   = MIN -MU +U(I) 
         DO 110 IX=F_FIRST2(I),F_LAST2(I)
            J  = F_HEAD2(IX)
            IS = F_COST2(IX)
            VJ = H + IS + V(J)
            IF (VJ .LT. LC(J)) THEN
               LC(J)  = VJ
               LAB(J) = I
               IF (VJ .GT. 0) GOTO 110
C     
                  IF (VJ .EQ. MIN) THEN
                     II = ROW(J)
                     IF ( II .GT. 0) THEN 
                        IF (U(II) .EQ. 0) THEN
                           LT = MIN
                           LABT = II
                           GOTO 150
                        ENDIF
                     ENDIF
                  ENDIF
C
                  IF (Q(J) .EQ. 0) THEN 
                     NHP = NHP + 1
                     Q(J) = NHP
                  ENDIF
                  IQ = Q(J)
90                IQ2 = IQ/2
                  IF (IQ2 .LE. 0) GOTO 100
                  HP2 = HP(IQ2)
                  IF (LC(J) .GE. LC(HP2)) GOTO 100
                  HP(IQ) = HP2
                  Q(HP2) = IQ
                  IQ = IQ2
                  GOTO 90
100               HP(IQ) = J
                  Q(J) = IQ
               ENDIF
110        CONTINUE 
      ELSE
C
C UPDATING FROM A COLUMN TO ANOTHER COLUMN
C
         H = LC(J0)
         DO 140 L=UP,MAXCOL
            J = PERM(L)
            IF (LC(J) .EQ. MIN) GOTO 140
C     
C UPDATE ONLY TO ASSIGNED COLUMNS
C
            I = ROW(J)
            IF (I .LE. 0) GOTO 140
            VJ = H + V(J)
            IF (VJ .LT. LC(J)) THEN
               LC(J)  = VJ
               LAB(J) = -J0
               IF (VJ .GT. 0) GOTO 140
C
               IF (VJ .EQ. MIN) THEN
                  IF (U(I) .EQ. 0) THEN
                     LT = MIN
                     LABT = I
                     GOTO 150
                  ENDIF
               ENDIF
               IF (Q(J) .EQ. 0) THEN
                  NHP = NHP + 1
                  Q(J) = NHP
               ENDIF
               IQ = Q(J)
120            IQ2 = IQ/2
               IF (IQ2 .LE. 0) GOTO 130
               HP2 = HP(IQ2)
               IF (LC(J) .GE. LC(HP2)) GOTO 130
               HP(IQ) = HP2
               Q(HP2) = IQ
               IQ = IQ2
               GOTO 120
130            HP(IQ) = J
               Q(J) = IQ
            ENDIF
140     CONTINUE
      ENDIF
      IF (NHP .NE. 0) GOTO 30
C
C UPDATING OF THE DUAL VARIABLES
C
150   CONTINUE
      ALFA = INF
      DO J=1,MAXCOL
         IF ( ROW(J) .LE. 0 ) THEN
            IF ( LC(J) .LT. ALFA) ALFA = LC(J)
         ENDIF
      ENDDO
      IF ( LT .GT. 0 )   LT = 0
      IF ( LT .LT. ALFA) ALFA = LT
C
C ELIMINATION OF POSSIBLE DUMMY ARCS
C
      IF (LT .LT. 0) THEN
         I = LABT
         J = COL(I)
160      IF (MU-U(I)-V(J) .EQ. BIGM) NAF = NAF-1
         I = LAB(J)
         IF (I .LT. 0) THEN
            J = -I
            I = LAB(J)
         ENDIF
         J = COL(I)
         IF (MAXROW .NE. I) GOTO 160
      ENDIF
C
      MU = MU + ALFA - LT
      DO J=1,MAXCOL
         I  = ROW(J)
         IF (LC(J) .LE. LT) THEN
            IF (I .GT. 0) THEN
                V(J) = V(J) + ALFA - LC(J)
                U(I) = U(I) - LT + LC(J)
            ELSE
                V(J) = 0
            ENDIF
         ELSE
            IF ( I.GT. 0) THEN
               V(J) = V(J) + ALFA -LT
            ELSE
               V(J) = 0
            ENDIF
         ENDIF
      ENDDO
      U(MAXROW) = -LT
      Z         = Z + LT
      IF ( LT .LT. 0) THEN
C
C UPDATE THE PRIMAL SOLUTION
C
         I     = LABT
         J     = COL(I)
         COL(I)= 0
170      I     = LAB(J)
         IF (I .LT. 0) THEN
            ROW(J) = 0
            J = -I
            I = LAB(J)
         ENDIF
         ROW(J)= I
         IQ     = J
         J     = COL(I)
         COL(I)= IQ
         IF (MAXROW .NE. I) GOTO 170
      ENDIF
C
      RETURN
      END
C================================================================================
      SUBROUTINE CHEAP(B_COST,B_TAIL,B_FIRST,F_COST2,F_HEAD2,
     +           F_FIRST2,F_LAST2,MAXROW,B_COST2,B_TAIL2,B_FIRST2,
     +           B_LAST2,MAXCOL,ROW,COL,MU,U,V,Z,INVR,C,INVC,
     +           NUMC,BIGM,RID,NAF,DELTA,INF)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)      
C
C THIS SUBROUTINE ADDS THE (MAXCOL+1)-TH COLUMN TO THE SPARSE MATRIX AND
C REOPTIMIZES THE CURRENT SOLUTION
C
C INPUT***************************************************************************
C B_COST,B_TAIL,B_FIRST = BACKWARD STAR OF THE ORIGINAL MATRIX
C F_COST2,F_HEAD2,F_FIRST2,F_LAST2 = FORWARD STAR OF THE REDUCED MATRIX
C MAXROW = NUMBER OF ROWS IN THE REDUCED MATRIX
C B_COST2,B_TAIL2,B_FIRST2,B_LAST2 = BACKWARD STAR OF THE REDUCED MATRIX
C MAXCOL   = NUMBER OF COLUMNS IN THE REDUCED MATRIX
C ROW,COL  = CURRENT PRIMAL SOLUTION
C MU, U, V = CURRENT DUAL VALUES
C C(IQ)    = INDEX OF THE COLUMN OF THE ORIGINAL MATRIX CORRESPONDING TO COLUMN IQ
C            OF THE NEW SPARSE MATRIX
C INVC()   = COLUMNS 'INVERTED' INDEX
C INVR()   = ROWS' INVERTED INDEX: ROWS( INVR(H) ) = H
C NUMC     = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX IN THE COLUMNS
C            OF THE REDUCED MATRIX
C BIGM     = VALUE OF THE DUMMY ELEMENTS
C RID      = NUMBER OF ELEMENTS OF THE ORIGINAL MATRIX WHICH HAVE NOT BEEN INSERTED
C NAF      = NUMBER OF DUMMY ELEMENTS
C DELTA    = UPPER BOUND TO THE VALUE OF AN ELEMENT TO BE INSERTED IN SOLUTION
C INF      = A VERY LARGE NUMBER
C
C OUTPUT***************************************************************************
C UPDATED VALUES OF THE REDUCED MATRIX AND OF THE CURRENT PRIMAL AND DUAL SOLUTIONS
C
      INTEGER*4 B_COST(ORDS),B_FIRST(ORDN+1),
     + F_COST2(ORDS),F_FIRST2(ORDM),F_LAST2(ORDM),
     + B_COST2(ORDS),B_FIRST2(ORDN),B_LAST2(ORDN),MU,U(ORDM),V(ORDN),
     + Z,MAXROW,MAXCOL,C(ORDM),INVC(ORDN),INVR(ORDM),ROW(ORDN),
     + COL(ORDM),INF,RID,NUMC,BIGM,NAF,DELTA
      INTEGER*4 LAB(ORDM),LR(ORDM),HP(ORDN),IPERM(ORDM),PERM(ORDM),
     +          Q(ORDN)
      INTEGER*2 B_TAIL(ORDS),F_HEAD2(ORDS),B_TAIL2(ORDS)
C
C LOCAL VARIABLES
C
      INTEGER ALFA,CHK,LRI,I,IA,IX,J,II,I0,IQ,LABT,LT,H,MIN,UI,JJ,IB,
     + IS,INDEX,NEL,NHP,HP1,HP2,HP3,IQ2,IN,UP,I1,L
C
      MAXCOL = MAXCOL + 1
      DO I=1,MAXROW
         LR(I) = INF
         HP(I) = 0
         Q(I) = 0
         PERM(I) = I
         IPERM(I) = I
      ENDDO
      NHP = 0
      J      = C(MAXCOL)
      INVC(J)= MAXCOL
C
C CHK = 0 IF THE CURRENT SOLUTION REMAINS OPTIMAL
C     = -1 OTHERWHISE
C
      CHK = 0
      IB = NUMC
      NEL = B_FIRST(J+1) - B_FIRST(J)
      NUMC = NUMC + NEL
      B_FIRST2(MAXCOL) = IB+1
      DO IX=B_FIRST(J),B_FIRST(J+1)-1
         I  = B_TAIL(IX)
         II = INVR(I)
         IF (II .LE. MAXROW) THEN
            IA = B_COST(IX)
            IF (IA.LE.DELTA) THEN
               IB = IB+1
               B_LAST2(MAXCOL) = IB
               B_TAIL2(IB) = II
               B_COST2(IB) = IA
               INDEX = F_LAST2(II) + 1
               IF (INDEX .EQ. 1) INDEX = F_FIRST2(II)
               F_LAST2(II) = INDEX
               F_HEAD2(INDEX) = MAXCOL
               F_COST2(INDEX) = IA
               LR(II)  = IA - MU + U(II)
               LAB(II) = MAXCOL
               IF (IA - MU+ U(II) .LT. 0) THEN
                   CHK = -1
                   IF (Q(II) .EQ. 0) THEN
                      NHP = NHP + 1
                      Q(II) = NHP
                   ENDIF
                   IQ = Q(II)
10                 IQ2 = IQ/2
                   IF (IQ2 .LE. 0) GOTO 20
                   HP2 = HP(IQ2)
                   IF (LR(II) .GE. LR(HP2)) GOTO 20
                   HP(IQ) = HP2
                   Q(HP2) = IQ
                   IQ = IQ2
                   GOTO 10
20                 HP(IQ) = II
                   Q(II) = IQ
               ENDIF
            ELSE
               RID = RID + 1
            ENDIF
         ENDIF
      ENDDO
      ROW(MAXCOL) = 0
      V(MAXCOL)   = 0
      IF (CHK .EQ. 0) RETURN
C
C ALTERNATING PATH TO REOPTIMIZE THE SOLUTION (IF NECESSARY)
C
      UP = 1
      LT = INF
30    I0  = HP(1)
      MIN = LR(I0)
      IF (MIN .GE. LT) GOTO 150
      IN = IPERM(I0)
      I1 = PERM(UP)
      IPERM(I0) = UP
      IPERM(I1) = IN
      PERM(UP) = I0
      PERM(IN) = I1
      UP = UP + 1
      Q(I0)= 0
      NHP = NHP - 1
      IF (NHP .EQ. 0) GOTO 80
      HP1 = HP(NHP+1)
      LRI = LR(HP1)
      IQ = 1
40    IQ2 = IQ*2
      IF (IQ2-NHP) 50,60,70
50    HP2 = HP(IQ2)
      HP3 = HP(IQ2+1)
      IF (LR(HP2) .LE. LR(HP3)) GOTO 60
      HP2 = HP3
      IQ2 = IQ2 + 1
60    HP2 = HP(IQ2)
      IF (LRI .LT. LR(HP2)) GOTO 70
      HP(IQ) = HP2
      Q(HP2) = IQ
      IQ = IQ2
      GOTO 40
70    HP(IQ) = HP1
      Q(HP1) = IQ
C     
80    J = COL(I0)
      IF ( J .GT. 0) THEN 
         H = MIN + V(J)
         IF (H .LT. LT) THEN 
            LT   = H
            LABT = J
            IF (V(J) .EQ. 0) GOTO 150
        ENDIF  
        H   = MIN -MU +V(J) 
        DO 110 IX=B_FIRST2(J),B_LAST2(J)
           I  = B_TAIL2(IX)
           IS = B_COST2(IX)
           UI = H + IS + U(I)
           IF (UI .LT. LR(I)) THEN
              LR(I)  = UI
              LAB(I) = J
              IF (UI .GT. 0) GOTO 110
C          
              IF (UI .EQ. MIN) THEN
                 JJ = COL(I)
                 IF ( JJ .GT. 0) THEN 
                    IF (V(JJ) .EQ. 0) THEN
                       LT = MIN
                       LABT = JJ
                       GOTO 150
                    ENDIF
                 ENDIF
              ENDIF
C
              IF (Q(I) .EQ. 0) THEN
                 NHP = NHP + 1
                 Q(I) = NHP
              ENDIF
              IQ = Q(I)
90            IQ2 = IQ/2
              IF (IQ2 .LE. 0) GOTO 100
              HP2 = HP(IQ2)
              IF (LR(I) .GE. LR(HP2)) GOTO 100
              HP(IQ) = HP2
              Q(HP2) = IQ
              IQ = IQ2
              GOTO 90
100           HP(IQ) = I
              Q(I) = IQ
           ENDIF
110    CONTINUE 
      ELSE
C
C UPDATING FROM A COLUMN TO ANOTHER COLUMN
C
        H = LR(I0)
        DO 140 L=UP,MAXROW
           I = PERM(L)
           IF (LR(I) .EQ. MIN) GOTO 140
C       
C UPDATE ONLY TO ASSIGNED COLUMNS
C
           J = COL(I)
           IF (J .LE. 0) GOTO 140
           UI = H + U(I)
           IF (UI .LT. LR(I)) THEN
              LR(I)  = UI
              LAB(I) = -I0
              IF (UI .GT. 0) GOTO 140
C                                    
              IF (UI .EQ. MIN) THEN
                 IF (V(J) .EQ. 0) THEN
                    LT = MIN
                    LABT = J
                    GOTO 150
                 ENDIF
              ENDIF
C
              IF (Q(I) .EQ. 0) THEN
                 NHP = NHP + 1
                 Q(I) = NHP
              ENDIF
              IQ = Q(I)
120           IQ2 = IQ/2
              IF (IQ2 .LE. 0) GOTO 130
              HP2 = HP(IQ2)
              IF (LR(I) .GE. LR(HP2)) GOTO 130
              HP(IQ) = HP2
              Q(HP2) = IQ
              IQ = IQ2
              GOTO 120
130           HP(IQ) = I
              Q(I) = IQ
           ENDIF
140     CONTINUE
      ENDIF
      IF (NHP .NE. 0) GOTO 30
C
C UPDATING OF THE DUAL VARIABLES
C
150   CONTINUE
      ALFA = INF
      DO I=1,MAXROW
         IF ( COL(I) .LE. 0 ) THEN
            IF ( LR(I) .LT. ALFA) ALFA = LR(I)
         ENDIF
      ENDDO
      IF ( LT .GT. 0 )   LT = 0
      IF ( LT .LT. ALFA) ALFA = LT
C
C ELIMINATION OF EVENTUAL DUMMY ARCS
C
      IF (LT .LT. 0) THEN
         J = LABT
         I = ROW(J)
160      IF (MU-U(I)-V(J) .EQ. BIGM) NAF = NAF-1
         J = LAB(I)
         IF (J .LT. 0) THEN
            I = -J
            J = LAB(I)
         ENDIF
         I = ROW(J)
         IF (MAXCOL .NE. J) GOTO 160
      ENDIF
C
      MU = MU + ALFA - LT
      DO I=1,MAXROW
         J  = COL(I)
         IF (LR(I) .LE. LT) THEN
            IF (J .GT. 0) THEN
               U(I) = U(I) + ALFA - LR(I)
               V(J) = V(J) - LT + LR(I)
            ELSE
               U(I) = 0
            ENDIF
         ELSE
            IF ( J .GT. 0) THEN
               U(I) = U(I) + ALFA -LT
            ELSE
               U(I) = 0
            ENDIF
         ENDIF
      ENDDO
      V(MAXCOL) = -LT
      Z         = Z + LT
      IF ( LT .LT. 0) THEN
C
C UPDATE THE PRIMAL SOLUTION
C
         J     = LABT
         I     = ROW(J)
         ROW(J)= 0
170      J     = LAB(I)
         IF (J .LT. 0) THEN
            COL(I) = 0
            I = -J
            J = LAB(I)
         ENDIF
         COL(I)= J
         IQ     = I
         I     = ROW(J)
         ROW(J)= IQ
         IF (MAXCOL .NE. J) GOTO 170
      ENDIF
C
      RETURN
      END
C================================================================================
      SUBROUTINE DEFNET(M,N,K,F_HEAD,F_FIRST,NA)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
      PARAMETER (ORD2=ORDM+ORDN+3)
      PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C  THIS SUBROUTINE BUILDS THE NETWORK USED TO FIND AN INITAL PRIMAL 
C  SOLUTION BY SOLVING A MAX FLOW PROBLEM ON IT THROUGH DINIC'S ALGORITHM
C
      INTEGER   M,N,K
      INTEGER*2 F_HEAD(ORDS)
      INTEGER*4 F_FIRST(ORDM+1)
C
      INTEGER   I,J,I1,NA,NN
C
      INTEGER*4 DNFAPT(ORD2),DNCAP(ORD4)
      INTEGER*2 DNFADJ(ORD4)
      COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP
C
      NA = 0
      DO I=1,M
         DNFAPT(I) = F_FIRST(I)
         DO I1=F_FIRST(I),F_FIRST(I+1)-1
            NA = NA + 1
            DNFADJ(NA) = M + F_HEAD(I1)
            DNCAP(NA) = 1
         ENDDO
      ENDDO
      DO J=1,N
         DNFAPT(M+J) = NA + 1
         NA = NA + 1
         DNFADJ(NA) = M + N + 3
         DNCAP(NA) = 1
      ENDDO
      DNFAPT(M+N+1) = NA + 1
      DO I=1,M
         NA = NA + 1
         DNFADJ(NA) = I
         DNCAP(NA) = 1
      ENDDO 
      DNFAPT(M+N+2) = NA + 1
      NA = NA + 1
      DNFADJ(NA) = M + N + 1
      DNCAP(NA) = K
      NN = M + N + 3
      DNFAPT(M+N+3) = NA + 1
      DNFAPT(NN+1) = NA + 1
C
      RETURN
      END
C================================================================================
      SUBROUTINE GREEDY(M,N,B_COST,B_TAIL,B_FIRST,INF,K,ROW,COL,R,
     +           PMINR,PSMINR,C,PMINC,PSMINC,HZ,HNASS,FLAG,MINRC,
     +           PTMINR,PTMINC)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
C
C INPUT*******************************************************************
C M          = NUMBER OF ROWS
C N          = NUMBER OF COLUMNS
C B_FIRST(J) = FIRST ENTRY OF B_COST(), B_TAIL() CORRESPONDING TO COLUMN J
C B_COST(H)  = VALUE OF THE H-TH ENTRY OF MATRIX W
C B_TAIL(H)  = ROW INDEX OF THE H-TH ENTRY OF MATRIX W
C INF        = SUFFICIENTLY LARGE VALUE
C K          = CARDINALITY OF THE REQUIRED MATCHING
C ROW(J)     = ROW ASSIGNED TO COLUMN J IN THE PARTIAL SOLUTION
C COL(I)     = COLUMN ASSIGNED TO THE ROW I IN THE PARTIAL SOLUTION
C R(H)       = H-TH ROW BY INCREASING VALUES OF THE MINIMUM OF THE ROW
C PMINR(I)   = (I,PMINR(I))IS THE ELEMENT CORRESPONDING TO THE FIRST 
C              MINIMUM OF THE ROW
C PSMINR(I)  = (I,PSMINR(I)) IS THE ELEMENT CORRESPONDIG TO THE SECOND
C              MINIMUM OF THE ROW
C PTMINR(I)  = (I,PTMINR(I)) IS THE ELEMENT CORRESPONDIG TO THE THIRD
C              MINIMUM OF THE ROW
C C(H)       = H-TH COL BY INCREASING VALUES OF THE MINIMUM OF THE COL.
C PMINC(I)   = (I,PMINC(I))IS THE ELEMENT CORRESPONDING TO THE FIRST 
C              MINIMUM OF THE COLUMN
C PSMINC(I)  = (I,PSMINC(I)) IS THE ELEMENT CORRESPONDIG TO THE SECOND
C              MINIMUM OF THE COLUMN
C PTMINC(I)  = (I,PTMINC(I)) IS THE ELEMENT CORRESPONDIG TO THE SECOND
C              MINIMUM OF THE COLUMN
C HZ         = VALUE OF THE PARTIAL SOLUTION
C HNASS      = NUMBER OF ASSIGNMENTS IN THE PARTIAL SOLUTION
C FLAG       = 1 THE PARTIAL SOLUTION WAS OBTAINED BY MEANS OF COLUMN REDUCTION
C            = 2 THE PARTIAL SOLUTION WAS OBTAINED BY MEANS OF ROWS REDUCTION
C MINRC      = IF FLAG .EQ. 1:
C              COLUMNS C(1),...,C(MINRC) ARE ASSIGNED IN THE PARTIAL SOLUTION
C              AND C(MINRC+1) CANNOT BE ASSIGNED
C            = IF FLAG .EQ. 2:
C              ROWS R(1),...,R(MINRC) ARE ASSIGNED IN THE PARTIAL SOLUTION
C              AND R(MINRC+1) CANNOT BE ASSIGNED
C
C OUTPUT***********************************************************************
C ROW(J)     = IF > 0: ROW ASSIGNED TO COLUMN J IN THE PARTIAL SOLUTION
C              IF < 0: ROW ASSIGNED TO COLUMN J BY THE HEURISTIC
C COL(I)     = IF > 0: COLUMN ASSIGNED TO ROW I IN THE PARTIAL SOLUTION
C              IF < 0: COLUMN ASSIGNED TO ROW I BY THE HEURISTIC
C HZ         = COST OF THE HEURISTIC SOLUTION
C HNASS      = (.LE. K) NUMBER OF ASSIGNMENTS IN THE HEURISTIC SOLUTION
C
      INTEGER*4 M,N,B_COST(ORDS),B_FIRST(ORDN+1),
     +          INF,K,COL(ORDM),ROW(ORDN),
     +          R(ORDM),PMINR(ORDM),PSMINR(ORDM),C(ORDN),
     +          PMINC(ORDN),PSMINC(ORDN),HZ,HNASS,FLAG,MINRC,
     +          PTMINR(ORDM),PTMINC(ORDN)
      INTEGER*2 B_TAIL(ORDS)
C
C LOCAL VARIABLES
C
      INTEGER*4 I,II,J,JJ,IX,IA,IB,I0,J0
C
C EXECUTE STEPS A. AND B. OF ALGORITHM GREEDY IN [2]
C
      IF (FLAG .EQ. 1) THEN
         II = 1
         JJ = MINRC + 2
         IF ( JJ .GT. N ) GO TO 90
      ELSE
         JJ = 1
         II = MINRC + 2
         IF ( II .GT. M ) GO TO 60
      ENDIF
10    I = R(II)
      IF ( COL(I) .EQ. 0 .AND. ROW(PMINR(I)) .EQ. 0 ) GO TO 20
      II = II + 1
      IF ( II .GT. M ) GO TO 60
      GO TO 10
20    J = C(JJ)
      IF ( ROW(J) .EQ. 0 .AND. COL(PMINC(J)) .EQ. 0 ) GO TO 30
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 90
      GO TO 20
30    HNASS = HNASS + 1
      J0 = PMINR(I)
      DO IX=B_FIRST(J0),B_FIRST(J0+1)-1
         IF (B_TAIL(IX) .EQ. I) THEN
            IA = B_COST(IX)
            GOTO 40
         ENDIF
      ENDDO
40    I0 = PMINC(J)
      DO IX=B_FIRST(J),B_FIRST(J+1)-1
         IF (B_TAIL(IX) .EQ. I0) THEN
            IB = B_COST(IX)
            GOTO 50
         ENDIF
      ENDDO
50    IF ( IA.LT. IB) THEN
          COL(I) = - PMINR(I)
          ROW(PMINR(I)) = - I
          HZ = HZ + IA
          IF ( HNASS .EQ. K ) GO TO 120
          II = II + 1
          IF ( II .GT. M ) GO TO 60
      ELSE
          COL(PMINC(J)) = - J
          ROW(J) = - PMINC(J)
          HZ = HZ + IB
          IF ( HNASS .EQ. K ) GO TO 120
          JJ = JJ + 1
          IF ( JJ .GT. N ) GO TO 90
      ENDIF
      GO TO 10
C
C COMPLETE THE HEURISTIC SOLUTION WITH ONLY MINC
C
60    J = C(JJ)
      IF ( ROW(J) .EQ. 0 .AND. COL(PMINC(J)) .EQ. 0 ) GO TO 70
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 120
      GO TO 60
70    HNASS = HNASS + 1
      COL(PMINC(J)) = - J
      ROW(J) = - PMINC(J)
      I0 = PMINC(J)
      DO IX=B_FIRST(J),B_FIRST(J+1)-1
         IF (B_TAIL(IX) .EQ. I0) THEN
            IB = B_COST(IX)
            GOTO 80
         ENDIF
      ENDDO
80    HZ = HZ + IB
      IF ( HNASS .EQ. K ) GO TO 120
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 120
      GO TO 60
C
C COMPLETE THE HEURISTIC SOLUTION WITH ONLY MINR
C
90    I = R(II)
      IF ( COL(I) .EQ. 0 .AND. ROW(PMINR(I)) .EQ. 0 ) GO TO 100
      II = II + 1
      IF ( II .GT. M ) GO TO 120
      GO TO 90
100   HNASS = HNASS + 1
      COL(I) = - PMINR(I)
      ROW(PMINR(I)) = - I
      J0 = PMINR(I)
      DO IX=B_FIRST(J0),B_FIRST(J0+1)-1
         IF (B_TAIL(IX) .EQ. I) THEN
            IA = B_COST(IX)
            GOTO 110
         ENDIF
      ENDDO
110   HZ = HZ + IA
      IF ( HNASS .EQ. K ) GO TO 120
      II = II +1
      IF ( II .GT. M ) GO TO 120
      GO TO 90
120   CONTINUE
      IF ( HNASS .EQ. K ) GOTO 240
C
C EXECUTE STEP C. ALGORITHM GREEDY IN [2]
C
      IF (FLAG .EQ. 1) THEN
         II = 1
         JJ = MINRC + 2
         IF ( JJ .GT. N ) GO TO 210
      ELSE
         JJ = 1
         II = MINRC + 2
         IF ( II .GT. M ) GO TO 180
      ENDIF
      J0 = -1
      I0 = -1
130   I = R(II)
      IF ( COL(I) .EQ. 0 ) THEN
         IF ( PSMINR(I) .NE. 0 .AND. ROW(PSMINR(I)) .EQ. 0 ) THEN
            J0 = PSMINR(I)
         ELSE
            IF ( PTMINR(I) .NE. 0 .AND.
     +           ROW(PTMINR(I)) .EQ. 0 ) J0 = PTMINR(I)
         ENDIF
      ENDIF
      IF ( J0 .NE. -1 ) GOTO 140
      II = II + 1
      IF ( II .GT. M ) GO TO 180
      GO TO 130
140   J = C(JJ)
      IF ( ROW(J) .EQ. 0 ) THEN
         IF ( PSMINC(J) .NE. 0 .AND. COL(PSMINC(J)) .EQ. 0 ) THEN
            I0 = PSMINC(J)
         ELSE
            IF ( PTMINC(J) .NE. 0 .AND.
     +           COL(PTMINC(J)) .EQ. 0 ) I0 = PTMINC(J)
         ENDIF
      ENDIF
      IF ( I0 .NE. -1 ) GOTO 150
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 210
      GO TO 140
150   HNASS = HNASS + 1
      DO IX=B_FIRST(J0),B_FIRST(J0+1)-1
         IF (B_TAIL(IX) .EQ. I) THEN
            IA = B_COST(IX)
            GOTO 160
         ENDIF
      ENDDO
160   CONTINUE
      DO IX=B_FIRST(J),B_FIRST(J+1)-1
         IF (B_TAIL(IX) .EQ. I0) THEN
            IB = B_COST(IX)
            GOTO 170
         ENDIF
      ENDDO
170   IF ( IA .LT. IB ) THEN
          COL(I) = -J0
          ROW(J0) = - I
          HZ = HZ + IA
          IF ( HNASS .EQ. K ) GO TO 240
          II = II + 1
          IF ( II .GT. M ) GO TO 180
      ELSE
          COL(I0) = - J
          ROW(J) = - I0
          HZ = HZ + IB
          IF ( HNASS .EQ. K ) GO TO 240
          JJ = JJ + 1
          IF ( JJ .GT. N ) GO TO 210
      ENDIF
      I0 = -1
      J0 = -1
      GO TO 130
C
C COMPLETE THE HEURISTIC SOLUTION WITH ONLY SMINC,TMINC
C
180   J = C(JJ)
      I0 = -1
      IF ( ROW(J) .EQ. 0 ) THEN
         IF ( PSMINC(J) .NE. 0 .AND. COL(PSMINC(J)) .EQ. 0) THEN
            I0 = PSMINC(J)
         ELSE
            IF ( PTMINC(J) .NE. 0 .AND.
     +           COL(PTMINC(J)) .EQ. 0 ) I0 = PTMINC(J)
         ENDIF
      ENDIF
      IF (I0 .NE. -1 ) GOTO 190
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 240
      GO TO 180
190   HNASS = HNASS + 1
      COL(I0) = - J
      ROW(J) = - I0
      DO IX=B_FIRST(J),B_FIRST(J+1)-1
         IF (B_TAIL(IX) .EQ. I0) THEN
            IB = B_COST(IX)
            GOTO 200
         ENDIF
      ENDDO
200   HZ = HZ + IB
      IF ( HNASS .EQ. K ) GO TO 240
      JJ = JJ + 1
      IF ( JJ .GT. N ) GO TO 240
      GO TO 180
C
C COMPLETE THE HEURISTIC SOLUTION WITH ONLY SMINR,TMINR
C
210   I = R(II)
      J0 = -1
      IF ( COL(I) .EQ. 0 ) THEN
         IF ( PSMINR(I) .NE. 0 .AND. ROW(PSMINR(I)) .EQ. 0) THEN
            J0 = PSMINR(I)
         ELSE
            IF ( PTMINR(I) .NE. 0 .AND.
     +           ROW(PTMINR(I)) .EQ. 0 ) J0 = PTMINR(I)
         ENDIF
      ENDIF
      IF (J0 .NE. -1 ) GOTO 220
      II = II + 1
      IF ( II .GT. M ) GO TO 240
      GO TO 210
220   HNASS = HNASS + 1
      COL(I) = - J0
      ROW(J0) = - I
      DO IX=B_FIRST(J0),B_FIRST(J0+1)-1
         IF (B_TAIL(IX) .EQ. I) THEN
            IA = B_COST(IX)
            GOTO 110
         ENDIF
      ENDDO
230   HZ = HZ + IA
      IF ( HNASS .EQ. K ) GO TO 240
      II = II +1
      IF ( II .GT. M ) GO TO 240
      GO TO 90
240   CONTINUE
      IF ( HNASS .LT. K ) HZ = INF  ! STEP D. ALGORITHM GREEDY IN [2]
      RETURN
      END
C================================================================================
      SUBROUTINE SPJV_PUNT(N,F_COST,F_HEAD,B_COST,B_TAIL,
     +    F_FIRST,B_FIRST,
     +    X,Y,U,V,Z,LAB,D,FREE,COL,ICOL,INF,PR,IPR,PC,IPC)
      IMPLICIT NONE
C
C THIS SUBROUTINE SOLVES THE SPARSE LINEAR ASSIGNMENT PROBLEM
C ACCORDING TO JONKER & VOLGENANT, UVA.
C
C THE SUBROUTINE SOLVES THE SQUARE ASSIGNMENT ON MATRIX C() CONSIDERING
C ONLY THE N < ORD ROWS AND COLUMNS POINTED BY VECTORS PR() AND PC()
C THE N ROWS AND COLUMNS ARE THOSE FIXED IN THE PREPROCESSING
C
C INPUT **************************************************************************
C N      = NUMBER OF ROWS AND COLUMNS TO BE CONSIDERED (= K)
C F_COST,F_HEAD,F_FIRST = FORWARD STAR OF THE ORIGINAL MATRIX
C PR(I)  = I-TH ROW TO BE CONSIDERED
C PC(J)  = J-TH COL TO BE CONSIDERED
C IPR(I) = ROWS' INVERTED INDEX: PR( IPR(H) ) = H
C IPC(J) = COLUMNS' INVERTED INDEX: PC( IPC(H) ) = H
C
C OUTPUT PARAMETERS***************************************************************
C X(J)=I : I-TH FIXED ROW (IN PR) ASSIGNED TO J-TH FIXED COLUMN (IN PC)
C Y(I)=J : J-TH FIXED COLUMN (IN PC) ASSIGNED TO I-TH FIXED ROW (IN PR)
C U      = DUAL ROW VARIABLES (THE DUAL OPTIMAL SOLUTION ON THE ORIGINAL MATRIX
C          IS NOT GUARANTEED)
C V      = DUAL COLUMN VARIABLES (THE DUAL OPTIMAL SOLUTION ON THE ORIGINAL MATRIX
C          IS NOT GUARANTEED)
C Z      = VALUE OF OPTIMAL SOLUTION
C
      INTEGER N,F_COST(1),B_COST(1),X(1),Y(1),U(1),V(1),Z,
     +        F_FIRST(1), B_FIRST(1),ICOL(1),IPR(1),IPC(1)
      INTEGER*2 F_HEAD(1),B_TAIL(1)
      INTEGER LAB(1),D(1),FREE(1),COL(1),PR(1),PC(1),INF
C
C LOCAL VARIABLES
C
      INTEGER H,UI,U0,DI,LOW,MIN,CNT,VP,LAST,IX,JJ,IC
      INTEGER PRI,PCJ,PCJ0,COST,I,J,I0,J0,I1,L,L0,IQ,IVP
C
C INITIALIZATION
      DO 10 JJ=1,N
        X(JJ)=0
        J = PC(JJ)
        IPC(J) = JJ
        I = PR(JJ)
        IPR(I) = JJ
10    CONTINUE
      DO 30 I0=1,N
        I = N-I0+1
        UI = INF
        PRI = PR(I)
        DO 20 IX=F_FIRST(PRI),F_FIRST(PRI+1)-1
           PCJ = F_HEAD(IX)
           IF (IPC(PCJ) .EQ. 0) GOTO 20
           IF (F_COST(IX) .LT. UI) THEN   
               UI = F_COST(IX)
               J0 = IPC(PCJ)
           ENDIF
20      CONTINUE
        U(PRI) = UI
        COL(I) = I
        ICOL(I) = I
        IF (X(J0).EQ.0) THEN
          X(J0) = I
          Y(I)  = J0
        ELSE
          X(J0) = -ABS(X(J0))
          Y(I)  = 0
        END IF
30    CONTINUE
      L=0
      DO 50 J=1,N
        IF (X(J).EQ.0) THEN
          L=L+1
          FREE(L)=J
          GOTO 50
        END IF
        IF (X(J).LT.0) THEN
          X(J)=-X(J)
        ELSE
          PCJ = PC(J)
          IF (B_FIRST(J) .EQ. B_FIRST(J+1)-1) GO TO 50
          I1=X(J)
          MIN=INF
          DO 40 IX=B_FIRST(PCJ),B_FIRST(PCJ+1)-1
             PRI = B_TAIL(IX)
             IF (IPR(PRI) .EQ. I1) GOTO 40
             IF (B_COST(IX)-U(PRI) .LT. MIN) MIN = B_COST(IX)-U(PRI)
40        CONTINUE
          U(PR(I1)) = U(PR(I1))-MIN
        END IF
50    CONTINUE
C
C IMPROVE THE INITIAL SOLUTION
C
      CNT=0
      IF (L.EQ.0) GOTO 220
60    L0=L
      IQ=1
      L=0
70    J=FREE(IQ)
      PCJ= PC(J)
      IQ=IQ+1
      IF (B_FIRST(PCJ) .EQ. B_FIRST(PCJ+1)-1) THEN
         L = L + 1
         FREE(L) = J
         GO TO 100
      ENDIF
      U0= INF
      UI=INF
      DO 80 IX=B_FIRST(PCJ),B_FIRST(PCJ+1)-1
         PRI = B_TAIL(IX)
         IF (IPR(PRI) .EQ. 0) GOTO 80
         H = B_COST(IX) - U(PRI)
         IF (H .LT. UI) THEN
            IF (H .GE. U0) THEN
               UI = H
               I1 = IPR(PRI)
            ELSE
               UI = U0
               U0 = H
               I1 = I0
               I0 = IPR(PRI)
            ENDIF
         ENDIF
80    CONTINUE
      J0=Y(I0)
      IF (U0.LT.UI) THEN
        U(PR(I0))=U(PR(I0))-UI+U0
      ELSE
        IF (J0.EQ.0) GOTO 90
        I0=I1
        J0=Y(I1)
      END IF
      IF (J0.EQ.0) GOTO 90
      IF (U0.LT.UI) THEN
        IQ=IQ - 1
        FREE(IQ)=J0
      ELSE
        L=L+1
        FREE(L)=J0
      END IF
90    X(J)=I0
      Y(I0)=J
100   IF (IQ.LE.L0) GOTO 70
      CNT=CNT+1
      IF ((L.GT.0).AND.(CNT.LT.2)) GOTO 60
C
C AUGMENTATION PART
C
      L0=L
      DO 210 L=1,L0
         DO I=1,N
            D(I) = INF
         ENDDO
         J0=FREE(L)
         PCJ0 = PC(J0)
         DO 110 IX=B_FIRST(PCJ0),B_FIRST(PCJ0+1)-1
            PRI = B_TAIL(IX)
            I = IPR(PRI)
            IF (I. EQ. 0) GOTO 110
            D(I) = B_COST(IX) - U(PRI)
            LAB(I) = J0
110      CONTINUE
         VP   = 1
         LOW  = 1
120      LAST = LOW-1
         MIN = D(COL(VP))
         VP=VP+1
         DO 130 IQ=VP,N
           I = COL(IQ)
           DI= D(I)
           IF (DI.LE.MIN) THEN
              IF (DI.LT.MIN) THEN
                 MIN=DI
                 VP=LOW
              END IF
              IVP = COL(VP)
              ICOL(IVP) = IQ
              ICOL(I) = VP
              COL(IQ)=IVP
              COL(VP)=I
              VP=VP+1
           END IF
130      CONTINUE
         DO 140 H=LOW,VP-1
            I=COL(H)
            IF (Y(I).EQ.0) GOTO 180
140      CONTINUE
150      I0 = COL(LOW)
         LOW= LOW+1
         J  = Y(I0)
         PCJ = PC(J)
         DO IX=B_FIRST(PCJ),B_FIRST(PCJ+1)-1
            IF (B_TAIL(IX) .EQ. PR(I0)) THEN
               IC = B_COST(IX)
               GOTO 160
            ENDIF
         ENDDO
160      H  = IC-U(PR(I0))-MIN
         DO 170 IX=B_FIRST(PCJ),B_FIRST(PCJ+1)-1
            PRI = B_TAIL(IX)
            I = IPR(PRI)
            IF (I .EQ. 0) GOTO 170
            IF (ICOL(I) .LT. VP) GOTO 170
            UI = B_COST(IX) - U(PRI) - H
            IF (UI .LT. D(I)) THEN
               D(I) = UI
               LAB(I) = J
               IF (UI .EQ. MIN) THEN
                  IF (Y(I) .EQ. 0) GOTO 180
                  IQ = ICOL(I)
                  IVP = COL(VP)
                  ICOL(IVP) = IQ
                  ICOL(I) = VP
                  COL(IQ) = IVP
                  COL(VP) = I
                  VP = VP + 1
               ENDIF
            ENDIF
170      CONTINUE
         IF (LOW.EQ.VP) GOTO 120
         GOTO 150
180      DO 190 IQ=1,LAST
            I0   = COL(IQ)
            U(PR(I0))= U(PR(I0))+D(I0)-MIN
190      CONTINUE
200      J   = LAB(I)
         Y(I)= J
         IQ   = I
         I   = X(J)
         X(J)= IQ
         IF (J0.NE.J) GOTO 200
210   CONTINUE
220   Z=0
      DO 240 J=1,N
         I = X(J)
         PCJ = PC(J)
         DO IX=B_FIRST(PCJ),B_FIRST(PCJ+1)-1
            IF (B_TAIL(IX) .EQ. PR(I)) THEN
               COST = B_COST(IX)
               GOTO 230
            ENDIF
         ENDDO      
230      V(PC(J)) = COST - U(PR(I))
         Z = Z + COST
240   CONTINUE
      RETURN
      END
C================================================================================
      SUBROUTINE SPJV_MOD(N,CC,KK,FIRST,LAST,X,Y,U,V,Z,LAB,D,
     +                FREE,TODO,OK,INF,BIGM,NAF)
      IMPLICIT NONE
C
C THIS SUBROUTINE SOLVES THE SPARSE LINEAR ASSIGNMENT PROBLEM
C ACCORDING TO JONKER & VOLGENANT, UVA.
C THE ALGORITHM IS MODIFIED SO AS TO MANAGE THE ADDITION OF DUMMY
C ELEMENTS IN ORDER TO HAVE ALWAYS AN ASSIGNMENT OF CARDINALITY K 
C (NEW STEP 3 IN [2])
C
C INPUT ********************************************************************
C N        = NUMBER OF ROWS AND COLUMNS (= K)
C FIRST(J) = FIRST ENTRY OF THE SPARSE MATRIX IN KK,CC
C LAST(J)  = LAST ENTRY OF THE SPARSE MATRIX IN KK,CC
C KK(FIRST(J),...,LAST(J)) = ROWS OF ELEMENTS OF COLUMN J
C CC(H)    = COST OF THE H-TH ELEMENTS OF THE SPARSE MATRIX
C BIGM     = VALUE OF THE DUMMY ARC NECESSARY TO HAVE A FEASIBLE PROBLEM
C
C OUTPUT********************************************************************
C X   = ROW ASSIGNED TO COL
C Y   = COL ASSIGNED TO ROW
C U   = DUAL ROW VARIABLE
C V   = DUAL COLUMN VARIABLE
C Z   = VALUE OF OPTIMAL SOLUTION
C NAF = NUMBER OF DUMMY ARCS USED
C
      INTEGER N,CC(1),FIRST(1),LAST(1),U(1),V(1),X(1),Y(1),Z
      INTEGER*2 KK(1)
      INTEGER LAB(1),D(1),FREE(1),TODO(1),INF
      INTEGER*4 BIGM,NAF
      LOGICAL OK(1)
C
C LOCAL VARIABLES
C
      INTEGER CNT,L0,T,T0,TD,V0,VI,DI,EL,H
      INTEGER NASS1,I,J,IQ,I0,L,MIN,I1,J0,LAJ,CNT1
C
C INITIALIZATION
C
      NAF = 0
      NASS1=0
      DO 10 I=1,N
      U(I)=INF
      Y(I) = 0
10    CONTINUE
      DO 30 J=1,N
      X(J)=0
      DO 20 T=FIRST(J),LAST(J)
      I=KK(T)
      IF (CC(T).LT.U(I)) THEN
        U(I)=CC(T)
        Y(I)=J
      END IF
20    CONTINUE
30    CONTINUE
      DO 40 I=1,N
      I0=N-I+1
      J=Y(I0)
      IF (J .EQ. 0) GOTO 40
      IF (X(J).NE.0) THEN
        X(J)=-ABS(X(J))
        Y(I0)=0
        NASS1=NASS1-1
      ELSE
        X(J)=I0
        NASS1=NASS1+1
      END IF
40    CONTINUE
      L=0
      DO 60 J=1,N
      IF (X(J).EQ.0) THEN
        L=L+1
        FREE(L)=J
        GOTO 60
      END IF
      IF (X(J).LT.0) THEN
        X(J)=-X(J)
      ELSE
        IF (FIRST(J) .EQ. LAST(J)) GOTO 60
        I1=X(J)
        MIN=INF
        DO 50 T=FIRST(J),LAST(J)
          I=KK(T)
          IF (I.EQ.I1) GOTO 50
          IF (CC(T)-U(I).LT.MIN) MIN=CC(T)-U(I)
50      CONTINUE
        U(I1)=U(I1)-MIN
      END IF
60    CONTINUE
C
C IMPROVE THE INITIAL SOLUTION
C
      CNT=0
      CNT1=0
      IF (L.EQ.0) GOTO 280
70    L0=L
      IQ=1
      L=0
80    J=FREE(IQ)
      CNT1=CNT1+1
      IF (CNT1.EQ.2*N) THEN
         IF (L+1.NE.IQ) THEN
            DO I=IQ,L0
               L=L+1
               FREE(L)=FREE(I)
            ENDDO
         ELSE
            L=L+L0-IQ+1
         ENDIF
         GOTO 120
      ENDIF
      IQ=IQ+1
C
C IF COLUMN J HAS ONLY ONE ELEMENT AND HAS NOT BEEN ASSIGNED 
C THEN STOP TRYING
C
      IF (FIRST(J) .EQ. LAST(J)) THEN
         L = L + 1
         FREE(L) = J
         GOTO 110
      ENDIF
      V0=INF
      I0 = 1
      VI=INF
      DO 90 T=FIRST(J),LAST(J)
      I=KK(T)
      H=CC(T)-U(I)
      IF (H.LT.VI) THEN
        IF (H.GE.V0) THEN
          VI=H
          I1=I
        ELSE
          VI=V0
          V0=H
          I1=I0
          I0=I
        END IF
      END IF
90    CONTINUE
      IF (V0.GE.INF) THEN
         L=L+1
         FREE(L)=J
         GOTO 110
      ENDIF
      J0=Y(I0)
      IF (V0.LT.VI) THEN
        U(I0)=U(I0)-VI+V0
      ELSE
        IF (J0.EQ.0) GOTO 100
        I0=I1
        J0=Y(I1)
      END IF
      IF (J0.EQ.0) GOTO 100
      IF (V0.LT.VI) THEN
        IQ=IQ-1
        FREE(IQ)=J0
        NASS1=NASS1-1
      ELSE
        L=L+1
        FREE(L)=J0
        NASS1=NASS1-1
      END IF
100   X(J)=I0
      Y(I0)=J
      NASS1=NASS1+1
110   IF (IQ.LE.L0) GOTO 80
      CNT=CNT+1
      IF ((L.GT.0).AND.(CNT.LT.2)) GOTO 70
C
C AUGMENTATION PART
C
120   CONTINUE
      L0=L
      DO 270 L=1,L0
      DO 130 I=1,N
      OK(I)=.FALSE.
      D(I)=INF
130   CONTINUE
      MIN=INF-1
      IQ = 0
      J0=FREE(L)
      TD=N
      DO 140 T=FIRST(J0),LAST(J0)
      I=KK(T)
      DI=CC(T)-U(I)
      D(I)=DI
      LAB(I)=J0
      IF (DI.LE.MIN) THEN
        IF (DI.LT.MIN) THEN
          MIN=DI
          IQ=1
          TODO(1)=I
        ELSE
          IQ=IQ+1
          TODO(IQ)=I
        END IF
      END IF
140   CONTINUE
      DO 150 H=1,IQ
      I=TODO(H)
      IF (Y(I).EQ.0) GOTO 260
      OK(I)=.TRUE.
150   CONTINUE
C
C REPEAT UNTIL A FREE ROW HAS BEEN FOUND
C
160   IF (IQ.EQ.0) GOTO 210
      I0=TODO(IQ)
      IQ=IQ-1
      J=Y(I0)
      TODO(TD)=I0
      TD=TD-1
      T0=FIRST(J)
      LAJ = LAST(J)
      T=T0-1
170   T=T+1
C
C PUT A DUMMY ELEMENT IN THE SOLUTION
C N.B. IT IS NOT NECESSARY TO STORE IT IN THE MATRIX
C 
      IF (T .GT. LAJ) THEN
         EL = BIGM
         GOTO 180
      ENDIF
      IF (KK(T).NE.I0) GOTO 170
      EL= CC(T)
180   H=EL-U(I0)-MIN
      DO 190 T=T0,LAJ
      I=KK(T)
      IF (.NOT. OK(I)) THEN
        VI=CC(T)-H-U(I)
        IF (VI.LT.D(I)) THEN
          D(I)=VI
          LAB(I)=J
          IF (VI.EQ.MIN) THEN
            IF (Y(I).EQ.0) GOTO 240
            IQ=IQ+1
            TODO(IQ)=I
            OK(I)=.TRUE.
          END IF
        END IF
      END IF
190   CONTINUE
      IF (IQ.NE.0) GOTO 160
      MIN=INF-1
      DO 200 I=1,N
      IF (D(I).LE.MIN) THEN
        IF (.NOT. OK(I)) THEN
          IF (D(I).LT.MIN) THEN
            MIN=D(I)
            IQ=1
            TODO(1)=I
          ELSE
            IQ=IQ+1
            TODO(IQ)=I
          END IF
        END IF
      END IF
200   CONTINUE
210   IF (IQ .EQ. 0) THEN
         NAF = NAF + 1
         I = 0
220      I = I + 1
         IF (Y(I) .NE. 0) GOTO 220
         IF (U(I) .EQ. INF) U(I) = BIGM
         MIN = BIGM - U(I)
         LAB(I) = J0
         GOTO 240
      ENDIF
      DO 230 I0=1,IQ
      I=TODO(I0)
      IF (Y(I).EQ.0) GOTO 240
      OK(I)=.TRUE.
230   CONTINUE
      GOTO 160
240   IF (MIN.EQ.0) GOTO 260
      DO 250 IQ=TD+1,N
      I0=TODO(IQ)
      IF (D(I0) .GT. MIN) GOTO 250
      U(I0)=U(I0)+D(I0)-MIN
250   CONTINUE
260   J=LAB(I)
      Y(I)=J
      IQ=I
      I=X(J)
      X(J)=IQ
      IF (J0.NE.J) GOTO 260
270   CONTINUE
280   Z=0
      DO 310 J=1,N
      LAJ = LAST(J)
      I=X(J)
      T=FIRST(J)-1
290   T=T+1
      IF (T .GT. LAJ) THEN
         DI = BIGM
         GOTO 300
      ENDIF
      IF (KK(T).NE.I) GOTO 290
      DI=CC(T)
300   V(J)=DI-U(I)
      Z=Z+DI
310   CONTINUE
C
      RETURN
      END
C================================================================================
      SUBROUTINE LOWERB(M,N,K,FROW,FCOL,NSETFROW,NSETFCOL,
     + MINR,PMINR,MINC,PMINC,SMINR,SMINC,R,C,INF,HZ,LB)
      IMPLICIT NONE
      INTEGER*4 ORDM,ORDN,ORDS
      PARAMETER (ORDM=3000)
      PARAMETER (ORDN=3000)
      PARAMETER (ORDS=1000000)
C
C M        = NUMBER OF ROWS
C N        = NUMBER OF  COLUMNS
C K        = REQUIRED CARDINALITY
C COL(I)   = COLUMN ASSIGNED TO ROW I IN THE PARTIAL OPTIMAL SOLUTION
C ROW(J)   = ROW    ASSIGNED TO COL J IN THE PARTIAL OPTIMAL SOLUTION
C FROW(I)  = .TRUE. IF ROW I IS FIXED, .FALSE. OTHERWISE
C FCOL(J)  = .TRUE. IF COLUMN J IS FIXED, .FALSE. OTHERWISE
C NSETFROW = NUMBER OF FIXED ROWS
C NSETFCOL = NUMBER OF FIXED COLUMNS
C MINR(I)  = VALUE OF THE FIRST MINIMUM OF ROW I
C PMINR(I) = COLUMN CORRESPONDING TO THE FIRST MINIMUM OF ROW I
C MINC(J)  = VALUE OF THE FIRST MINIMUM OF COLUMN J
C PMINC(J) = ROW    CORRESPONDING TO THE FIRST MINIMUM OF COLUMN J
C SMINR(I) = VALUE OF THE SECOND MINIMUM OF ROW I
C SMINC(J) = VALUE OF THE SECOND MINIMUM OF COLUMN J
C R(H)     = H-TH ROW, ORDERED BY INCREASING MINIMA
C C(H)     = H-TH COLUMN, ORDERED BY INCREASING MINIMA
C INF      = INFINITY
C HZ       = VALUE OF THE BEST APPROXIMATE SOLUTION
C OUTPUT****************************************************************
C LB       = LOWER BOUND
C
      INTEGER*4 M,N,K,MINR(M),MINC(N),PMINR(M),
     + PMINC(N),R(M),C(N),NSETFROW,NSETFCOL,HZ,SMINR(M),SMINC(N),
     + INF,LB
      LOGICAL FROW(N),FCOL(N)
C
C LOCAL VARIABLES
C
      INTEGER*4 PC(ORDN),PR(ORDM),INC(ORDN),INR(ORDM)
      INTEGER*4 I,J,IQ,II,JJ,KK,SUMR,SUMC,SUM1,SUM,MIN,DELTA
      EQUIVALENCE (PC(1),INC(1))
      EQUIVALENCE (PR(1),INR(1))
C
C ROWS
C PC(J)  =  0  :NO ROW HAS THE MINIMUM IN COLUMN J
C        = -I  :ONLY ROW I HAS MINIMUM IN COLUMN J
C        =  I  :ROW I IS THE FIRST OF THE CHAIN OF ROWS HAVING 
C               THE MINIMUM IN COLUMN J
C INR(I) =      NEXT ROW OF THE CHAIN 
C
      LB = 0
      SUMR = 0
      DO I=1,M
         INR(I) = 0
      ENDDO
      DO J=1,N
         PC(J) = 0
      ENDDO
C
C BASIC BOUND
C
      IQ  = NSETFROW
      KK = 0
      DO II = 1,M
         I = R(II)
         IF ( FROW(I) ) THEN
               SUMR = SUMR + MINR(I)
               KK = KK + 1
               IF (KK .EQ. NSETFROW .AND. IQ .EQ. K) GOTO 10
               J = PMINR(I)
               IF (PC(J) .EQ. 0) THEN
                  PC(J) = -I
                  INR(I) = 0
               ELSE
                  INR(I) = IABS(PC(J))
                  PC(J)  = I
               ENDIF            
         ELSE
            IF ( IQ .LT. K) THEN
               SUMR = SUMR + MINR(I)
               IQ = IQ + 1
               IF (KK .EQ. NSETFROW .AND. IQ .EQ. K) GOTO 10
            ENDIF
         ENDIF
      ENDDO
C
10    IF (LB .LT. SUMR) LB = SUMR
      IF (LB .EQ. HZ) GOTO 70
C
C IMPROVEMENT
C
      DO JJ = 1,N
         IF ( PC(JJ) .GT. 0) THEN
            I = PC(JJ)
            SUM = SMINR(I)
            SUM1 = MINR(I)
20          I = INR(I)
            IF (I .GT. 0) THEN
               SUM = SUM + SMINR(I)
               SUM1 = SUM1 + MINR(I)
               GOTO 20
            ENDIF
C
            MIN = INF
            I = PC(JJ)
            DELTA = SUM - SMINR(I) + MINR(I)
            IF (MIN .GT. DELTA) MIN = DELTA
30          I = INR(I)
            IF (I .GT. 0) THEN
               DELTA = SUM - SMINR(I) + MINR(I)
               IF (MIN .GT. DELTA) MIN = DELTA 
               GOTO 30
            ENDIF
C
            SUMR = SUMR + MIN - SUM1
            IF (LB .LT. SUMR) LB = SUMR
            IF (LB .EQ. HZ) GOTO 70
         ENDIF
      ENDDO
C
C COLUMNS
C PR(I)  =  0  :NO COLUMN HAS THE MINIMUM IN ROW I
C        = -J  :ONLY COLUMN J HAS MINIMUM IN ROW I
C        =  J  :COLUMN J IS THE FIRST OF THE CHAIN OF COLUMNS HAVING THE MINIMUM
C               IN ROW J
C INC(J) =     NEXT COLUMN OF THE CHAIN 
C
      DO I=1,M
         PR(I) = 0
      ENDDO
      DO J=1,N
         INC(J) = 0
      ENDDO
C
      SUMC = 0
C
C BASIC BOUND
C
      IQ  = NSETFCOL
      KK = 0
      DO JJ = 1,N
         J = C(JJ)
         IF (FCOL(J)) THEN
            SUMC = SUMC + MINC(J)
            KK = KK + 1
            IF (KK .EQ. NSETFCOL .AND. IQ .EQ. K) GOTO 40
            I = PMINC(J)
            IF (PR(I) .EQ. 0) THEN
               PR(I) = -J
               INC(J) = 0
            ELSE
               INC(J) = IABS(PR(I))
               PR(I)  = J
            ENDIF            
         ELSE
            IF ( IQ .LT. K) THEN
               SUMC = SUMC + MINC(J)
               IQ = IQ + 1
               IF (KK .EQ. NSETFCOL .AND. IQ .EQ. K) GOTO 40
            ENDIF
         ENDIF
      ENDDO
C
40    IF (LB .LT. SUMC) LB = SUMC
      IF (LB .EQ. HZ) GOTO 70
C
C IMPROVEMENT
C
      DO II = 1,M
         IF ( PR(II) .GT. 0) THEN
            J = PR(II)
            SUM = SMINC(J)
            SUM1 = MINC(J)
50          J = INC(J)
            IF (J .GT. 0) THEN
               SUM = SUM + SMINC(J)
               SUM1 = SUM1 + MINC(J)
               GOTO 50
            ENDIF
C
            MIN = INF
            J = PR(II)
            DELTA = SUM - SMINC(J) + MINC(J)
            IF (MIN .GT. DELTA) MIN = DELTA
60          J = INC(J)
            IF (J .GT. 0) THEN
               DELTA = SUM - SMINC(J) + MINC(J)
               IF (MIN .GT. DELTA) MIN = DELTA
               GOTO 60
            ENDIF
C
            SUMC = SUMC + MIN - SUM1
            IF (LB .LT. SUMC) LB = SUMC
            IF (LB .EQ. HZ) GOTO 70
         ENDIF
      ENDDO
C
70    RETURN
      END
C================================================================================
      SUBROUTINE SORTI(N,A,V,JDA)
C
C SORT THE INTEGER ARRAY A BY INCREASING VALUES (DERIVED FROM
C SUBROUTINE SORTZV OF THE C.E.R.N. LIBRARY).
C
C JDA           = LENGTH OF ARRAY A;
C N             = NUMBER OF ELEMENTS OF A TO BE SORTED;
C V(I) (INPUT)  = POINTER TO THE I-TH ELEMENT TO BE SORTED;
C V(I) (OUTPUT) = POINTER TO THE I-TH ELEMENT OF THE SORTED ARRAY.
C
C ON RETURN, ARRAY A IS UNCHANGED.
C
      INTEGER*4 V(N),IU(20),IL(20)
      INTEGER*4 A(JDA),T
      II = 1
      JJ = N
      IF ( N .LE. 1 ) GOTO 120
      M = 1
      I = II
      J = JJ
10    IF ( I .GE. J ) GO TO 80
20    IQ = I
      IJ = (J + I)/2
      IV = V(IJ)
      T = A(IV)
      IQI = V(I)
      IF ( A(IQI) .LE. T ) GO TO 30
      V(IJ) = IQI
      V(I) = IV
      IV = V(IJ)
      T = A(IV)
30    L = J
      IQI = V(J)
      IF ( A(IQI) .GE. T ) GO TO 50
      V(IJ) = IQI
      V(J) = IV
      IV = V(IJ)
      T = A(IV)
      IQI = V(I)
      IF ( A(IQI) .LE. T ) GO TO 50
      V(IJ) = IQI
      V(I) = IV
      IV = V(IJ)
      T = A(IV)
      GO TO 50
40    V(L) = V(IQ)
      V(IQ) = IVT
50    L = L - 1
      IQI = V(L)
      IF ( A(IQI) .GT. T ) GO TO 50
      IVT = IQI
60    IQ = IQ + 1
      IQI = V(IQ)
      IF ( A(IQI) .LT. T ) GO TO 60
      IF ( IQ .LE. L ) GO TO 40
      IF ( L - I .LE. J - IQ ) GO TO 70
      IL(M) = I
      IU(M) = L
      I = IQ
      M = M + 1
      GO TO 90
70    IL(M) = IQ
      IU(M) = J
      J = L
      M = M + 1
      GO TO 90
80    M = M - 1
      IF ( M .EQ. 0 ) GOTO 120
      I = IL(M)
      J = IU(M)
90    IF ( J - I .GE. II ) GO TO 20
      IF ( I .EQ. II ) GO TO 10
      I = I - 1
100   I = I + 1
      IF ( I .EQ. J ) GO TO 80
      IV = V(I+1)
      T = A(IV)
      IQI = V(I)
      IF ( A(IQI) .LE. T ) GO TO 100
      IQ = I
110   V(IQ+1) = V(IQ)
      IQ = IQ - 1
      IQI = V(IQ)
      IF ( T .LT. A(IQI) ) GO TO 110
      V(IQ+1) = IV
      GO TO 100
120   END
C================================================================================
C
C ALL THE SUBROUTINES FROM HERE TO THE END IMPLEMENT THE DINIC'S ALGORITHM FOR
C THE MAX FLOW PROBLEM AND THE CODE BELONGS TO:
C D. GOLDFARB AND M.D. GRIGORIADIS, "A COMPUTATIONAL COMPARISON OF THE DINIC
C AND SIMPLEX METHODS FOR MAXIMUM FLOW", IN: B. SIMEONE ET AL. EDS., FORTRAN CODES
C FOR NETWORK OPTIMIZATION, ANN. OPER. RES. 13, (BALZER, BASEL, 1988) 83-103
C THE AUTHORS HAVE MADE NO MODIFICATION TO THE ORIGINAL CODE (BUT THE SIZES OF 
C VECTORS) AND THE READER IS REFERRED TO THE ABOVE REFERENCE
C
C================================================================================
C
	SUBROUTINE DNSUB(NODES,NARCS,ISOURC,ISINK,MAXFLO,
     *                 NUMAUG,NUMSTG,NNCUT,NACUT,ELTIM,IRETN)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C************START DNSUB STANDARD USER XFACE DECLARATION BLOCK*********
C
C	USER MUST INCLUDE THIS BLOCK IN	HIS/HER	CALLING	PROGRAM.
C
C   THESE DECLARATIONS ACCOUNT FOR THE ENTIRE ARRAY STORAGE USED BY
C   THE	DNSUB SUBROUTINES, INCLUDING WHAT IS NECESSARY TO STORE	THE
C   INPUT DATA.	IT AMOUNTS TO 6*NODES +	5*ARCS WORDS, BUT IT MAY BE
C   STATED MORE	PRECISELY AS FOLLOWS:
C
C   1) FOR SOLVING PROBLEMS WITH UP TO 32,765 NODES AND	2,147,483,647
C   ARCS, AND WITH FLOW	VALUES OF UP TO	2,147,483,647, IT SUFFICES
C   TO DECLARE ARRAYS AS SHOWN UNDER 'STD ARRANGEMENT' IN THE
C   TABLE BELOW.  THIS ARRANGEMENT IS THE ONE USED IN THE RELEASE
C   VERSION OF THE DNSUB SUBROUTINES.
C
C   2) IN ORDER	TO HANDLE PROBLEMS WITH	MORE THAN 32,765 NODES,	ALL
C   INTEGER*2 DECLARATIONS BELOW MUST BE CHANGED TO INTEGER*4 AND
C   THE	DNSUB SUBROUTINES AND USER CALLING PROGRAMS MUST BE COMPILED
C   AND	LINKED.	(SEE ALTERNATE 1 BELOW.)
C
C   3) FOR SOLVING ONLY	PROBLEMS WITH AT MOST 32,765 NODES, 32,767
C   ARCS, AND WITH FLOWS NOT EXCEEDING 32,767, ALL INTEGER*4
C   DECLARATIONS BELOW MAY BE CHANGED TO INTEGER*2, AND	THE DNSUB
C   SUBROUTINES	 AND USER CALLING PROGRAMS MUST	BE COMPILED AND
C   LINKED. (SEE ALTERNATE 2 BELOW.)  IT IS ALSO NECESSARY TO
C   ACTIVATE THE STATEMENT DNIBIG=32767	in subroutine dnsub, and to
C   include 'implicit INTEGER*2 (i-n)' statements in each program
C   unit.  THIS	ARRANGEMENT, WHICH USES	THE LEAST AMOUNT OF MEMORY
C   AND	IS THE FASTEST when executed on	16-BIT PROCESSORS, IS ONLY
C   USEFUL FOR A SET OF	VERY SPECIAL APPLICATIONS.  ITS	general
C   use	is not recommended.  OTHER ALTERNATES ARE ALSO POSSIBLE.
C
C  -----------------------------------------------------------------
C
C		  STD ARRANGEMENT    ALTERNATE 1      ALTERNATE	2
C						   (not	recommended)
C		  ---------------  ---------------  ----------------
C  ARRAY   LENGTH TYPE EQV'D TYPE  TYPE EQV'D TYPE  TYPE EQV'D TYPE
C  ------  -----    ----  -------  ---- ----- ----  ---- ----- -----
C  DNFAPT    N      4               4                2
C  DNFADJ    A      2               4                2
C  DNCAP     A      4               4                2
C  DNBAPT    N      4               4                2
C  DNBADJ    A      2 DNFROM  2     4 DNFROM  4      2 DNFROM   2
C  DNFLOW    A      4 DNTO    2     4 DNTO    4      2 DNTO     2
C  DNBTOF    A      4 DNGCAP  4     4 DNGCAP  4      2 DNGCAP   2
C  DNPTRF    N      4               4                2
C  DNPTRB    N      4               4                2
C  DNLIST    N      2               4                2
C  DNFLAB    N      4 DNDIST  2     4 DNDIST  4      2 DNDIST   2
C  -----------------------------------------------------------------
C
C  TOTAL BYTES:     22*N + 16*A     24*N + 20*A       12*N + 10*A
C  (N=NODES, A=ARCS)
C  -----------------------------------------------------------------
C
C DIMENSION THESE TO AT LEAST MAX NUMBER OF NODES + 2:
        INTEGER*2   DNLIST(ORD2),DNDIST(ORD2)
        INTEGER*4   DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *            DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT LEAST MAX NUMBER OF ARCS + 1:
        INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
        INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
        COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
        COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
        COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
        INTEGER  DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *  ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
        INTEGER  DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
        COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *               DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
        EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *              (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C************END DNSUB STANDARD USER XFACE DECLARATION BLOCK***********
C***********************************************************************
C
C CALLING CONDITIONS:
C       USER CALLABLE WITH INPUT DATA AS FOLLOWS.
C INPUT:
C   SCALARS (IN CALLING SEQUENCE; ALL INTEGER*4):
C       NODES:  NUMBER OF NODES (INCLUDING SOURCE AND SINK)
C       NARCS:  NUMBER OF ARCS
C       ISOURC: NODE NUMBER FOR SOURCE
C       ISINK:  NODE NUMBER FOR SINK
C
C   ARRAYS (IN COMMON):
C       DNFAPT: (NODES+1)-LONG INTEGER*4 POINTER ARRAY FOR FORWARD ADJA-
C               CENCY LISTS (I.E. THE FORWARD ADJACENCY LIST OF A NODE I
C               IS THE SET OF ARCS
C                      {(I,DNFADJ(J)) :J=DNFAPT(I),...,DNFAPT(I+1)-1 }
C               NOTE: MUST HAVE DNFAPT(NODES+1) = NARCS+1.
C
C       DNFADJ: NARCS-LONG INTEGER*2 ARRAY GIVING THE LIST OF NODES IN
C               THE FORWARD ADJACENCY LIST DNFADJ(J) FOR EACH NODE J,
C               AS DESCRIBED ABOVE.
C
C       DNCAP:  NARCS-LONG INTEGER*4 ARRAY GIVING THE ARC CAPACITIES, IN
C               THE ORDER PRESCRIBED BY DNFAPT(.) AND DNFADJ(.). ALL ARC
C               CAPACITIES MUST BE GIVEN AS POSITIVE INTEGERS.
C
C   NOTE:       FOR TRANSFORMING UNORDERED ARC LISTS TO THE ABOVE INPUT
C               DATA STRUCTURE,USE SUBROUTINE DNFWD BEFORE CALLING DNSUB
C
C OUTPUT:
C
C   SCALARS (IN CALLING SEQUENCE):
C       MAXFLO: VALUE OF MAXIMUM FLOW (INTEGER*4).
C       NUMAUG: NUMBER OF FLOW AUGMENTATIONS (INTEGER*4).
C       NUMSTG: NUMBER OF STAGES (LAYERED NETWORKS CREATED; INTEGER*4).
C       NNCUT:  NUMBER OF NODES ON source SIDE OF FINAL CUT (INTEGER*4)
C       NACUT:  NUMBER OF SATURATED ARCS IN THE FINAL CUT (INTEGER*4)
C       ELTIM:  EXECUTION TIME, IN SECONDS (SEE NOTE 3 BELOW) (REAL*4).
C       IRETN:  NONZERO IF THERE ARE ERRORS IN INPUT DATA.
C
C   ARRAYS (IN COMMON:):
C       DNFAPT: THE ORIGINAL DNFAPT(.) WITH SOME OF ITS ELEMENTS
C               NEGATED TO MARK THE NODES THAT ARE ON THE SINK SIDE OF
C               THE MINIMUM CUT, I.E. DNFAPT(I)<0 IMPLIES THAT NODE I IS
C               ON THE SINK SIDE OF CUT, ELSE, IT IS ON THE source SIDE.
C
C       DNFADJ: THE ORIGINAL DNFADJ(.) WITH SOME OF ITS ELEMENTS NEGATED
C               TO MARK THOSE SATURATED ARCS THAT ARE IN THE MIN CUT
C               FOUND BY THE ALGORITHM, I.E.  DNFADJ(I)<0 IMPLIES THAT
C               THE I-TH ARC IN FORWARD ADJACENY ORDER IS IN THE CUT;
C               THE FOLLOWING CODE SEGMENT PRINTS THESE ARCS:
C
C                               DO 2 K=1,NODES
C                                  IBEG=IABS(DNFAPT(K))
C                                  IEND=IABS(DNFAPT(K+1))-1
C                                  DO 1 I=IBEG,IEND
C                                    IF(DNFADJ(I).LT.0)PRINT THE ARC
C                       1          CONTINUE
C                       2       CONTINUE
C
C               (SEE NOTE 1 BELOW).
C
C       DNCAP:  THE ORIGINAL CAPACITIES, UNALTERED.
C
C       DNFLOW: NARCS-LONG INTEGER*4 ARRAY WHICH GIVES FLOWS ON ARCS, IN
C               FORWARD ADJACENCY ORDER.
C
C       DNBAPT: (NODES+1)-LONG INTEGER*4 POINTER ARRAY FOR BACKWARD
C               ADJACENCY LISTS (I.E. THE FORWARD ADJACENCY LIST OF
C               NODE I IS THE SET OF ARCS
C                   {(DNBADJ(J),J): J=DNBAPT(I),...,DNBAPT(I+1)-1 }.
C               NOTE: WE MUST HAVE DNBAPT(NODES+1) = NARCS+1.
C
C       DNBADJ: NARCS-LONG INTEGER*2 ARRAY GIVING THE BACKWARD ADJACENCY
C               LIST OF EACH NODE J, ONE AFTER THE OTHER, AS DESCRIBED
C               ABOVE (ALSO SEE NOTE 2 BELOW).
C
C       DNBTOF: NARCS-LONG INTEGER*4 ARRAY GIVING THE BACKWARD ADJACENCY
C               TO FORWARD ADJACENCY MAPPING, I.E. THE J-TH ARC
C               IN THE BACKWARD ADJACENCY ORDER IS THE DNBTOF(J)-TH
C               ARC IN THE FORWARD ADJACENCY ORDER.
C
C***********************************************************************
C
C INTERNAL SCRATCH ARRAYS (IN COMMON):
C
C       DNLIST: (NODES+1)-LONG INTEGER*2 ARRAY USED AS FOLLOWS:
C               1) AS THE QUEUE IN BFS SEARCH FOR CONSTRUCTING THE
C                  LAYERED GRAPH (SUBR. DNBFS).
C               2) AS THE "PARENT" ARRAY IN DFS SEARCH OF LAYERED
C                  GRAPH (SUBR: DNDFS, DNPUSH).
C               3) AS SCRATCH IN CONSTRUCTING FWD ADJACENCIES
C                  (SUBR DNFWD), AND BWD ADJACENCIES (SUBR. DNSUB).
C
C       DNFLAB: NODES-LONG INTEGER*4 ARRAY USED IN DFS SEARCH OF
C               LAYERED GRAPH TO STORE THE FLOW LABEL FOR EACH NODE
C               (SEE ARRAY DNDIST() ).
C
C       DNDIST: NODES-LONG INTEGER*2 ARRAY USED IN CONSTRUCTING THE
C               LAYERED GRAPH BY BFS, STORING THE DISTANCE OF EACH
C               NODE FROM THE SOURCE; DNDIST(ISOURC)=0 (EQUIVALENCED
C               TO THE INTEGER*4 ARRAY DNFLAB() ).
C
C       DNPTRF: NODES-LONG INTEGER*4 STATUS AND POINTER ARRAY, I.E.
C               DNPTRF(K)=0 IF THERE ARE NO OPEN ARCS LEAVING NODE K
C                               IN THE REPRESENTATION OF THE LAYERED
C                               NETWORK.
C                        >0 THE ARC INDEX, IN 'FA' ORDER, THAT
C                               CORRESPONDS TO THE LAST ARC IN THE
C                               FWD ADJACENCY OF K WHICH IS SCANNED
C                               BACKWARD.
C
C       DNPTRB: NODES-LONG INTEGER*4 STATUS AND POINTER ARRAY, I.E.
C               DNPTRB(K)=0 IF THERE ARE NO OPEN ARCS ENTERING NODE K
C                               IN THE REPRESENTATION OF THE LAYERED
C                               NETWORK.
C                        >0 THE ARC INDEX (IN 'BA' ORDER) OF THE FIRST
C                               ARC IN THE BWD ADJACENCY OF K TO BE
C                               SCANNED IN LIFO ORDER.
C
C***********************************************************************
C
C NOTES:1) DNFADJ(J) IS INTERNALLY NEGATED TO MARK THE J-TH ARC (IN
C          FORWARD ADJACENCY ORDER) AS 'CLOSED'. BEFORE RETURNING
C          TO THE CALLING PGM, THESE MARKINGS ARE DISCARDED, AND SOME
C          ELEMENTS OF DNFADJ(.) ARE NEGATED TO REFLECT THE OUTPUT
C          SPECIFICATION DESCRIBED ABOVE.
C
C       2) DNBADJ(J) IS INTERNALLY NEGATED TO MARK THE J-TH ARC
C          (IN BACKWARD ADJACENCY ORDER) AS 'CLOSED'.
C
C       3) THE USER MUST PROVIDE A TIMING SUBROUTINE THAT RETURNS
C          THE CURRENT TIME IN VARIABLE ' T ', IN SECONDS.  SUCH
C          A SUBROUTINE IS INSTALLATION-DEPENDENT AND IS NOT GIVEN HERE.
C          IF THIS CANNOT BE DONE, PROVIDE THE FOLLOWING SUBROUTINE AND
C          FORGET THE EXECUTION TIMING INFORMATION:
C
C               SUBROUTINE GETIME(T)
C               REAL T
C               T=0.0
C               RETURN
C               END
C
C***********************************************************************
C
C       SUBROUTINES CALLED:     DNBFS, DNDFS, DNCUT, DNCLEA
C
C***********************************************************************
C
C   NAMES COLLECTIVELY RESERVED BY ALL DNSUB SUBROUTINES:
C
C   DN00, DN01, DN02, DN03, DN04, DN05, DN06, DN08, DN09, DN10, DN11,
C   DN12, DNARC, DNAUG, DNBADJ, DNBAPT, DNBFS, DNBTOF, DNCAP, DNCLEA,
C   DNCUT, DNDFS, DNDIST, DNFADJ, DNPUSH, DNFAPT, DNFLAB, DNFLOW,
C   DNFROM, DNFVA, DNFWD, DNGCAP, DNIBIG, DNLAUG, DNLFVA, DNLIST,
C   DNNODE, DNNOP2, DNOUT, DNPTRB, DNPTRF, DNSINK, DNSRCE, DNSTGE,
C   DNSUB, DNELTM, DNTO.
C
C**********************************************************************
C INITIALIZATION:
C
C        CALL GETTIME(TIMBEG)
        IRETN=0
        DNIBIG=2147483647
C       DNIBIG=32767
        DNARC=NARCS
        DNNODE=NODES
        DNNOP2=DNNODE+2
        DNSRCE=ISOURC
        DNSINK=ISINK
        DO 200 I4=1,DNARC
200     DNFLOW(I4)=0
C***********************************************************************
C
C USING THE FORWARD ADJACENCY ARRAYS DNFAPT(.) AND DNFADJ(.) (IN
C COMMON), CREATE THE BACKWARD ADJACENCY ARRAYS DNBAPT(.), DNBADJ(.)
C AND THE BACKWARD TO FORWARD ADJACENCY MAPPING DNBTOF(.).  DNLIST(.) IS
C USED HERE AS A SCRATCH ARRAY.
C
        NODP12=DNNODE+1
        DO 210 I2=1,NODP12
                DNLIST(I2)=0
210             DNBAPT(I2)=0
C TEMPORARILY STORE IN DNBAPT(.) NUMBER OF ARCS INTO EACH NODE:
        DO 230 I2=1,DNNODE
                IBEG4=DNFAPT(I2)
                IEND4=DNFAPT(I2+1)-1
                DO 220 I4=IBEG4,IEND4
                  JJDN=DNFADJ(I4)
220             DNBAPT(JJDN)=DNBAPT(JJDN)+1
230     CONTINUE
C CONSTRUCT DNBAPT(.):
        IHPI4=DNBAPT(1)
        DNBAPT(1)=1
        DO 240 I2=1,DNNODE
                IHPSV4=IHPI4+DNBAPT(I2)
                IHPI4=DNBAPT(I2+1)
240     DNBAPT(I2+1)=IHPSV4
C CONSTRUCT DNBADJ(.) AND DNBTOF(.):
        DO 260 I2=1,DNNODE
                IBEG4=DNFAPT(I2)
                IEND4=DNFAPT(I2+1)-1
                DO 250 I4=IBEG4,IEND4
                        IHEAD2=DNFADJ(I4)
                        IHPUT4=DNBAPT(IHEAD2)+DNLIST(IHEAD2)
                        IF(IHPUT4.LE.0)GOTO1900
                        DNBADJ(IHPUT4)=I2
                        DNBTOF(IHPUT4)=I4
250             DNLIST(IHEAD2)=DNLIST(IHEAD2)+1
260     CONTINUE
C
C***********************************************************************
C
        DNFVA=0
        DNAUG=0
        DNSTGE=1
C
C STAGE LOOP:
300     CONTINUE
C   FORM NEW LAYERED NETWORK W.R.T. CURRENT FEASIBLE FLOW.
                CALL DNBFS(LAYMAX)
                IF(LAYMAX.GT.0)GOTO1500
C   COMPUTE A MAXIMAL FLOW IN THIS LAYERED NET:
                CALL DNDFS
C   UPDATE FLOW VALUE:
                DNFVA=DNFVA+DNLFVA
                DNAUG=DNAUG+DNLAUG
                DNSTGE=DNSTGE+1
C   ERASE MARKINGS THAT DEFINE LAYERED NETWORK:
                CALL DNCLEA
        GOTO300
C***********************************************************************
C
C CURRENT FLOW IS MAXIMUM:
1500    MAXFLO=DNFVA
        NUMAUG=DNAUG
        NUMSTG=DNSTGE
        CALL DNCLEA
C MARK EXECUTION TIME:
C        CALL GETTIME(TIMEND)
        DNELTM=TIMEND-TIMBEG
        ELTIM=DNELTM
C
C MARK THE MIN CUT-SET IN DNFAPT(.) AND DNFADJ(.):
        CALL DNCUT(NNCUT,NACUT)
C
        RETURN
C
1900    IRETN=1
        RETURN
        END
****************************----------*********************************
        SUBROUTINE DNBFS(LAYMAX)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT LEAST MAX NUMBER OF NODES + 2:
        INTEGER*2   DNLIST(ORD2),DNDIST(ORD2)
        INTEGER*4   DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *              DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
        COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
        COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
        COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
        INTEGER  DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *  ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
        INTEGER  DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
        COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *               DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
        EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *              (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C*******************************************************************
C
C CALLING CONDITIONS:
C       INTERNAL CALL FROM DNSUB ONLY.
C INPUT ARRAYS:
C       DNFAPT, DNBAPT, DNFADJ, DNBADJ, DNBTOF, DNCAP, DNFLOW
C SCRATCH ARRAYS:
C       DNLIST
C OUTPUT:
C   IN CALLING SEQUENCE (INTEGER*2):
C       LAYMAX: 0 (LAYERED NET COMPLETE WITH SINK IN LAST LAYER.)
C               1 (LAST LAYER IS EMPTY; CURRENT FLOW IN ORIGINAL
C                  NETWORK IS MAXIMUM.)
C   IN COMMON (ARRAYS):
C       DNPTRF, DNPTRB, DNDIST
C***********************************************************************
C
C   THIS SUBROUTINE CONSTRUCTS THE LAYERED NETWORK OF THE RESIDUAL
C   NETWORK W.R.T. THE CURRENT FLOW.  THE RESIDUAL NETWORK IS
C   INFERRED, AND THE LAYERED NETWORK IS ONLY RECORDED BY TEMPORARY
C   MARKERS ON THE ORIGINAL GRAPH. THE SUBROUTINE IMPLEMENTS A
C   BREADTH-FIRST SEARCH ON THE RESIDUAL NETWORK, I.E.  NODES ARE
C   PROCESSED AS IN A QUEUE.  SUCH A SEARCH PARTITIONS THE NODES INTO
C   A SET OF 'LAYERS', ACCORDING TO THEIR CARDNALITY DISTANCE FROM
C   THE source. THE DNDIST(.) ARRAY IS USED TO RECORD THESE DISTANCES,
C   AND THUS DEFINES THE LAYERS (NOTE: DNDIST(ISOURC)=0).
C
C   IN SELECTING ARCS OF THE RESIDUAL NETWORK THAT ARE DIRECTED FROM
C   ONE LAYER TO THE NEXT, WE SHALL MARK THE ARCS IN THE ORIGINAL
C   NETWORK AS 'OPEN' OR 'CLOSED'.  A 'CLOSED' ARC (V,W) IN THE
C   LAYERED NETWORK IS EITHER AN ARC I=(V,W) THAT IS SATURATED IN THE
C   ORIGINAL NETWORK, OR AN ARC I=(W,V) OF THE ORIGINAL NETWORK WITH
C   ZERO FLOW. IN THE FORMER CASE, THE ABSENCE OF THIS ARC FROM THE
C   LAYERED NETWORK IS MARKED BY NEGATING DNFADJ(I), AND IN THE LATTER
C   CASE, BY NEGATING DNBADJ(I).  AN 'OPEN' ARC I=(V,W) FOR THE LAYERED
C   NETWORK EITHER HAS POSITIVE RESIDUAL CAPACITY, OR ARC (W,V) HAS
C   POSITIVE FLOW, IN THE ORIGINAL NETWORK. IF A NODE V HAS NO OPEN
C   ARCS IN THE LAYERED NETWORK THAT ARE IN THE FORWARD ADJACENCY LIST
C   OF V IN THE ORIGINAL NETWORK, THEN WE SET DNPTRF(V)=0; AND IF IT HAS
C   NO OPEN ARCS THAT ARE IN THE BACKWARD ADJACENCY LIST OF V IN THE
C   ORIGINAL NETWORK, WE SET DNPTRB(V)=0.
C
C   IN THIS BFS SEARCH WE ONLY USE OPEN ARCS THAT LEAD US TO NEW
C   NODES, OR TO NODES THAT HAVE ALREADY BEEN PLACED IN THE NEXT
C   LAYER. INITIALLY, THE QUEUE CONTAINS ONLY THE source.  AS THE
C   VERTICES ARE POPPED FROM THE QUEUE AND SCANNED, NEW NODES ARE
C   INJECTED INTO THE QUEUE. EVENTUALLY, EITHER THE SINK IS REACHED,
C   OR SOME VERTICES ARE NOT REACHABLE WITH OPEN ARCS FROM THE
C   CURRENT LAYER. IN THE FORMER CASE THE LAYERED NETWORK FOR THE
C   CURRENT STAGE IS COMPLETE, AND THUS A FLOW AUGMENTATION IS
C   POSSIBLE.  IN THE LATTER CASE, THE CURRENT FLOW ON THE ORIGINAL
C   NETWORK IS MAXIMUM, AND THE RUN TERMINATES.
C
C   THE QUEUE IS MAINTAINED IN THE LIST DNLIST(.) WITH TWO POINTERS,
C   AS SHOWN:
C                   -------------------------------------------
C    ARRAY              |    NODES TO BE SCANNED    |
C    DNLIST(.):  ...    |                           |
C                   -------------------------------------------
C                      ^                           ^
C                      |                           |
C                    QHEAD2                     QTAIL2
C*******************************************************************
C
C INITIALIZE:
                DO 10 I2=1,DNNODE
                        DNPTRF(I2)=0
                        DNPTRB(I2)=0
10              DNDIST(I2)=DNNOP2
                DNDIST(DNSRCE)=0
C PUT source INTO QUEUE:
                QHEAD2=0
                QTAIL2=1
                DNLIST(1)=DNSRCE
                MAXD2=DNNODE-1
C
C SCAN EACH NODE IN QUEUE:
C
C---------------
C
100     IF(QHEAD2.EQ.QTAIL2)GOTO2000
                QHEAD2=QHEAD2+1
C POP NODE IN FRONT OF QUEUE:
                K2=DNLIST(QHEAD2)
                K2DST=DNDIST(K2)
        IF(K2DST.GE.MAXD2)GOTO100
                K2DP1=K2DST+1
C
C  SCAN NODE K2, I.E. SEARCH OVER FWD ADJACENCY OF K2 AND FOR ARCS
C  (K2,J2) SUCH THAT J2 IS UNSCANNED AND J2 IS NOT IN THE QUEUE AND
C  ARC (K2,J2) HAS POSITIVE RESIDUAL CAPACITY IN ORIGINAL NETWORK.
C  'J2 UNSCANNED' IS CHECKED BY THE CONDITION: 'DNDIST(J2)>=DNDIST(K2)'.
C  'J2 NOT IN QUEUE' IS CHECKED BY THE CONDITION
C  'DNDIST(J2)=DNDIST(K2)+1 FOR	A SCANNED J2'.
C
                IBEG4=DNFAPT(K2)
                IEND4=DNFAPT(K2+1)-1
                IF(IBEG4.GT.IEND4)GOTO1000
                DO 400 I4=IBEG4,IEND4
                        J2=DNFADJ(I4)
                        IF(DNDIST(J2).LE.K2DST)GOTO300
                        IF(DNCAP(I4).LE.DNFLOW(I4))GOTO300
C (IMPLICITLY) MARK ARC (K2,J2) AS OPEN IN LAYERED NETWORK; ALSO,
C SAVE ITS INDEX IN DNPTRF(.):
                        DNPTRF(K2)=I4
                        IF(J2.NE.DNSINK)GOTO200
C J2=SINK:
                        MAXD2=K2DP1
                        DNDIST(J2)=MAXD2
                        GOTO400
C APPEND NODE J2 TO THE QUEUE, IF NOT ALREADY THERE:
200                     IF(DNDIST(J2).EQ.K2DP1)GOTO400
                        DNDIST(J2)=K2DP1
                        QTAIL2=QTAIL2+1
                        DNLIST(QTAIL2)=J2
                        GOTO400
C (EXPLICITLY) MARK ARC (K2,J2) AS CLOSED IN LAYERED NETWORK:
300                     DNFADJ(I4)=-DNFADJ(I4)
400             CONTINUE
C
C---------------
C
C IF SINK WAS REACHED, DON'T NEED TO SCAN BACKWARD ARCS	INTO ITS LAYER:
1000	IF(DNDIST(DNSINK).NE.DNNOP2)GOTO100
C
C ELSE,	CONTINUE SCAN OVER FORWARD ARCS	OUT OF K2 IN THE RESIDUAL
C NETWORK (WHICH ARE BACKWARD ARCS OF THE ORIGINAL NETWORK WITH
C POSITIVE FLOW). THIS CODE SEGMENT IS ANALOGOUS TO THE	ONE ABOVE:
C
		IBEG4=DNBAPT(K2)
		IEND4=DNBAPT(K2+1)-1
	IF(IBEG4.GT.IEND4)GOTO100
		DO 1400	I4=IBEG4,IEND4
			J2=DNBADJ(I4)
			IF(DNDIST(J2).LE.K2DST)GOTO1300
			JJDN=DNBTOF(I4)
			IF(DNFLOW(JJDN).LE.0)GOTO1300
			DNPTRB(K2)=I4
			IF(DNDIST(J2).EQ.K2DP1)GOTO1400
			DNDIST(J2)=K2DP1
			QTAIL2=QTAIL2+1
			DNLIST(QTAIL2)=J2
			GOTO1400
1300			DNBADJ(I4)=-DNBADJ(I4)
1400		CONTINUE
	GOTO100
C
C---------------
C
2000	LAYER2=LAYER2+1
	LAYP12=LAYER2+1
C HERE,	ALL NODES IN QUEUE HAVE	BEEN PROCESSED.
C IF THE SINK WAS REACHED, THE LAYERED NETWORK IS COMPLETE:
	LAYMAX=0
	IF(DNDIST(DNSINK).NE.DNNOP2)RETURN
C ELSE,	THE CURRENT FLOW IS MAXIMUM:
	LAYMAX=1
	RETURN
C
	END
	SUBROUTINE DNDFS
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		   DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C***********************************************************************
C
C  THIS	SUBROUTINE FINDS A MAXIMAL FLOW	IN THE LAYERED NETWORK.	IT
C  USES	A DEPTH-FIRST SEARCH STARTING FROM THE SOURCE AND RECORDS
C  THE DFS TREE	BY THE PREDECESSOR ARRAY DNLIST(.) AND FLOW LABEL
C  ARRAY DNFLAB(.). LAYERED NETWORK ARCS THAT ARE FORWARD IN THE
C  ORIGINAL NETWORK ARE	SEARCHED BEFORE	BACKWARD ARCS.	THE
C  FORWARD AND BACKWARD	ADJACENCIES OF A NODE K2 ARE SCANNED IN	LIFO
C  ORDER, STARTING FROM	THE LAST ARC THAT WAS VISITED DURING A
C  PREVIOUS SCAN OF K2.	THUS, WITH THE AID OF THE TWO POINTER LISTS
C  DNPTRF(.) AND DNPTRB(.), THE	ADJACENCY LISTS	OF THE ORIGINAL
C  NETWORK ARE SCANNED ONLY ONCE BY THIS SUBROUTINE. DNPTRF(K2)=0
C  IMPLIES THAT	NODE K2	HAS NO OUTGOING	UNSCANNED ARCS;	DNPTRB(K2)=0
C  IMPLIES THAT	NODE K2	HAS NO INCOMING	UNSCANNED ARCS;	A 'CLOSED'
C  NODE	IS DETECTED BY THE CONDITION DNPTRF(K2)=DNPTRB(K2)=0.
C
C***********************************************************************
C
C CALLING CONDITIONS:
C	INTERNAL CALL FROM DNSUB ONLY.
C INPUT	ARRAYS:
C	DNPTRB,	DNPTR, DNBTOF, DNCAP, DNFLOW, DNBADJ, DNFADJ
C OUTPUT ARRAYS:
C	DNPTRF,	DNPTRB,	DNLIST,	DNFLAB.
C***********************************************************************
C
C	SUBROUTINES CALLED:	DNPUSH
C
C***********************************************************************
C
	K2=DNSRCE
	DNLFVA=0
	DNLAUG=0
	DNFLAB(K2)=DNIBIG
C
C-----------
C
C SCAN NODE K2:
C
100	IF(DNPTRF(K2).EQ.0)GOTO1000
C FIND AN OPEN ARC FROM	K2 TO SOME NODE	J2: SCAN THE
C FORWARD ADJACENCY LIST OF K2,	STARTING FROM  DNPTRF(K2) AND
C PROCEEDING BACKWARD ON DNFADJ() TOWARD DNFAPT(K2):
		I4=DNPTRF(K2)
		IPT4=DNFAPT(K2)
300		J2=DNFADJ(I4)
			IF(J2.GE.0)GOTO600
400			I4=I4-1
		IF(I4.GE.IPT4)GOTO300
C NO OPEN ARCS OUT OF K2; MARK K2 AS CLOSED:
		DNPTRF(K2)=0
		GOTO1000
C AN OPEN ARC FOUND; MOVE POINTER DNPTRF() TO THIS ARC;
C EXTEND DFS TREE TO NODE J2:
600		DNPTRF(K2)=I4
		DNFLAB(J2)=DNFLAB(K2)
		IF(DNCAP(I4)-DNFLOW(I4).LT.DNFLAB(J2))
     *				DNFLAB(J2)=DNCAP(I4)-DNFLOW(I4)
		DNLIST(J2)=K2
		K2=J2
C IF NODE K2 (FORMERLY J2) IS THE SINK,	AUGMENT	FLOW; GET NEW K2:
		IF(K2.EQ.DNSINK)CALL DNPUSH(K2)
	GOTO100
C
C-----------
C
1000	IF(DNPTRB(K2).EQ.0)GOTO2000
C SCAN BACKWARD	ARCS INTO K2 TO	FIND AN	OPEN ARC:
		I4=DNPTRB(K2)
		IPT4=DNBAPT(K2)
1300		J2=DNBADJ(I4)
			IF(J2.GE.0)GOTO1600
			I4=I4-1
		IF(I4.GE.IPT4)GOTO1300
C NO OPEN ARC INTO K2:
		DNPTRB(K2)=0
		GOTO2000
C AN OPEN ARC FOUND; EXTEND DFS	TREE TO	NODE J2:
1600		DNPTRB(K2)=I4
		DNFLAB(J2)=DNFLAB(K2)
		JJDN=DNBTOF(I4)
		IFL4=DNFLOW(JJDN)
		IF(IFL4.LT.DNFLAB(J2))DNFLAB(J2)=IFL4
		DNLIST(J2)=-K2
		K2=J2
	GOTO100
C
C-----------
C
C K2 IS	'CLOSED' NODE; IF IT IS	THE source, THEN WE HAVE A MAXIMAL
C FLOW IN LAYERED NETWORK:
2000		IF(K2.EQ.DNSRCE)RETURN
C ELSE,	BACK UP	ONE NODE FROM K2 IN DFS	TREE:
		KP2=DNLIST(K2)
		K2=KP2
		IF(K2.LT.0)K2=-K2
		IF(KP2.GE.0)GOTO2200
		I4=DNPTRB(K2)
		DNBADJ(I4)=-DNBADJ(I4)
	GOTO100
2200		I4=DNPTRF(K2)
		DNFADJ(I4)=-DNFADJ(I4)
	GOTO100
C
C-----------
	END
	SUBROUTINE DNPUSH(K2)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		    DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C***********************************************************************
C
C AUGMENT FLOW ALONG FLOW AUGMENTING PATH DEFINED SUBROUTINE DNDFS, I.E.
C USING	THE PREDECESSOR	ARRAY DNLIST(.), START FROM THE	SINK AND
C TRAVERSE TO THE SOURCE ARC FLOWS AND FLOW LABELS BY AN AMOUNT	EQUAL
C TO DNFLAB(DNSINK). MARK SATURATED FORWARD (FROM S TO T) ARCS,	AND
C BACKWARD ARCS	HAVING ZERO FLOW, AS CLOSED.
C
C***********************************************************************
C
C CALLING CONDITIONS:
C	INTERNAL CALL FROM DNDFS ONLY.
C INPUT	ARRAYS:
C	DNFLAB,DNLIST,DNPTRB,DNPTRF,DNBTOF,DNCAP,DNFLOW,DNBADJ,DNFADJ
C OUTPUT ARRAYS:
C	DNFLAB,DNFLOW,DNBADJ,DNFADJ
C***********************************************************************
C
	KSAT2=0
	J2=DNSINK
	INCRE4=DNFLAB(J2)
	DNLFVA=DNLFVA+INCRE4
	DNLAUG=DNLAUG+1
C
100	K2=J2
		J2=DNLIST(K2)
		DNFLAB(K2)=DNFLAB(K2)-INCRE4
		IF(J2.GT.0)GOTO200
C
C DECREASE FLOW	ON BACKWARD ARC	(K2,J2):
		J2=-J2
		I4=DNPTRB(J2)
		II4=DNBTOF(I4)
		DNFLOW(II4)=DNFLOW(II4)-INCRE4
	IF(DNFLOW(II4).NE.0)GOTO100
C FLOW IS ZERO;	MARK ARC AS 'CLOSED' IN	LAYERED	NETWORK:
		DNBADJ(I4)=-DNBADJ(I4)
		KSAT2=J2
	GOTO100
C
C INCREASE FLOW	ON FORWARD ARC (J2,K2):
200		I4=DNPTRF(J2)
		DNFLOW(I4)=DNFLOW(I4)+INCRE4
		IF(DNCAP(I4).NE.DNFLOW(I4))GOTO300
C ARC IS NOW SATURATED;	MARK IS	AS 'CLOSED' IN LAYERED NET
		DNFADJ(I4)=-DNFADJ(I4)
		KSAT2=J2
300	IF(J2.NE.DNSRCE)GOTO100
C
C RETURN TO RESUME SEARCH FROM K2 (NODE	CLOSEST	TO source OF
C THE ARC CLOSED LAST)
	IF(KSAT2.EQ.0)STOP
	K2=KSAT2
	RETURN
	END
	SUBROUTINE DNFWD(NODES,NARCS,IRETN)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		    DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C***********************************************************************
C
C CALLING CONDITIONS:
C	USER CALLABLE; BEFORE CALLING DNSUB.
C INPUT:
C   IN CALLING SEQUENCE	(ALL INTEGER*4):
C	NODES=	NUMBER OF NODES, INCLUDING source AND SINK
C	NARCS=	NUMBER OF ARCS.
C
C   IN COMMON, PROVIDE THREE ARC LISTS WITH ARCS IN ANY	ORDER:
C	DNFROM=	THE LIST OF TAILS. (INTEGER*2)
C	DNTO=	THE LIST OF HEADS. (INTEGER*2)
C	DNGCAP=	THE LIST OF CAPACITIES.	(INTEGER*4).
C
C SCRATCH ARRAY	(IN COMMON):
C	DNLIST
C OUTPUT:
C   IN CALLING SEQUENCE	(INTEGER*4):
C	IRETN=0	 NO ERRORS;  =1	ERROR DETECTED,	CHECK INPUT.
C   IN COMMON, ARRAYS:
C	DNFAPT,DNFADJ,DNCAP (I.E. THE FWD ADJACENCY INPUT DATA STRUCTURE
C		 THAT IS REQUIRED AS INPUT TO SUBROUTINE DNSUB).
C
C***********************************************************************
C
C  USING THE THREE ARC LISTS DNFROM(.),	DNTO(.), DNGCAP(.), THIS
C  SUBROUTINE CONSTRUCTS THE FORWARD ADJACENCY ARRAYS DNFAPT(.),
C  DNFADJ(.), AND DNCAP(.), AS THE INPUT REQUIRED BY SUBROUTINE	DNSUB.
C
C***********************************************************************
C
C INITIALIZE:
	IRETN=0
	NODP1=NODES+1
	DO 100 I2=1,NODP1
		DNFAPT(I2)=0
100	DNLIST(I2)=0
C TEMPORARILY STORE IN DNFAPT(.) NUMBER	OF ARCS	OUT OF EACH NODE:
	DO 200 I4=1,NARCS
	  JJDN=DNFROM(I4)
200	DNFAPT(JJDN)=DNFAPT(JJDN)+1
C CONSTRUCT DNFAPT(.):
	ITPI4=DNFAPT(1)
	DNFAPT(1)=1
	DO 300 I2=1,NODES
		ITPSV4=ITPI4+DNFAPT(I2)
		ITPI4=DNFAPT(I2+1)
300	DNFAPT(I2+1)=ITPSV4
C CONSTRUCT DNFADJ(.) AND DNCAP(.):
	DO 400 I4=1,NARCS
		ITAIL2=DNFROM(I4)
		ITPUT4=DNFAPT(ITAIL2)+DNLIST(ITAIL2)
		IF(ITPUT4.LE.0)GOTO500
		DNFADJ(ITPUT4)=DNTO(I4)
		DNLIST(ITAIL2)=DNLIST(ITAIL2)+1
		DNCAP(ITPUT4)=DNGCAP(I4)
400	CONTINUE
	RETURN
500	IRETN=1
	RETURN
	END
	SUBROUTINE DNCLEA
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		    DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C*******************************************************************
C CALLING CONDITIONS:
C	INTERNAL CALL FROM DNSUB ONLY.
C INPUT	(IN COMMON):
C   SCALARS: DNARC
C   ARRAYS:  DNFADJ, DNBADJ
C OUTPUT:
C   ARRAYS (IN COMMON):	DNFADJ,	DNBADJ
C********************************************************************
C
C CLEAR	MARKINGS IN ARRAYS DNFADJ() AND	DNBADJ().:
	DO 100 I4=1,DNARC
	  IF(DNBADJ(I4).LT.0)DNBADJ(I4)=-DNBADJ(I4)
	  IF(DNFADJ(I4).LT.0)DNFADJ(I4)=-DNFADJ(I4)
100	CONTINUE
	RETURN
	END
	SUBROUTINE DNCUT(NNCUT,NACUT)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		    DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
C
C***********************************************************************
C CALLING CONDITIONS:
C	ONLY AFTER A SUCCESSFUL	RETURN FROM A CALL TO DNSUB.
C INPUT	ARRAYS:
C	DNFAPT,	DNFADJ,	DNCAP,DNFLOW, DNDIST
C OUTPUT ARRAYS:
C	DNFAPT,DNFADJ
C PASSED IN CALLING SEQUENCE:
C	NNCUT:	NUMBER OF NODES	ON source SIDE OF FINAL	CUT (INTEGER*4)
C	NACUT:	NUMBER OF ARCS IN THE FINAL CUT	(INTEGER*4)
C***********************************************************************
C
C NEGATE DNFAPT(.) FOR NODES ON	SINK SIDE OF CUT:
	NNCUT=0
	DO 1700	I2=1,DNNODE
		IF(DNDIST(I2).NE.DNNOP2)GOTO1650
		DNFAPT(I2)=-IABS(DNFAPT(I2))
		GOTO1700
1650		NNCUT=NNCUT+1
1700	CONTINUE
C NEGATE DNFADJ(.) FOR THOSE SATURATED ARCS IN MIN CUT:
	NACUT=0
	DO 1900	I2=1,DNNODE
		IF(DNFAPT(I2).LT.0)GOTO1900
C NODE I2 IS ON	source SIDE:
		IBEG4=IABS(DNFAPT(I2))
		IEND4=IABS(DNFAPT(I2+1))-1
		DO 1800	I4=IBEG4,IEND4
			JJDN=DNFADJ(I4)
			IF(DNFAPT(JJDN).GT.0)GOTO1800
C NODE DNFADJ(I4) ON SINK SIDE:
			IF(DNFLOW(I4).NE.DNCAP(I4))GOTO1800
			II2=DNFADJ(I4)
			DNFADJ(I4)=-IABS(II2)
			NACUT=NACUT+1
1800		CONTINUE
1900	CONTINUE
	RETURN
	END
	SUBROUTINE DNOUT(IUNREP)
        INTEGER*4 ORDM,ORDN,ORDS,ORD2,ORD4
        PARAMETER (ORDM=3000)
        PARAMETER (ORDN=3000)
        PARAMETER (ORDS=1000000)
        PARAMETER (ORD2=ORDM+ORDN+3)
        PARAMETER (ORD4=ORDS+ORDM+ORDN+1)
C
C
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF NODES + 2:
	INTEGER*2  DNLIST(ORD2),DNDIST(ORD2)
	INTEGER*4  DNFAPT(ORD2),DNPTRF(ORD2),DNBAPT(ORD2),
     *		    DNFLAB(ORD2),DNPTRB(ORD2)
C DIMENSION THESE TO AT	LEAST MAX NUMBER OF ARCS + 1:
      INTEGER*2 DNFADJ(ORD4),DNFROM(ORD4),DNTO(ORD4),DNBADJ(ORD4)
      INTEGER*4 DNCAP(ORD4),DNGCAP(ORD4),DNFLOW(ORD4),DNBTOF(ORD4)
C
	COMMON /DN01/DNFAPT/DN02/DNFADJ/DN03/DNCAP/DN04/DNFROM
	COMMON /DN05/DNFLOW/DN06/DNGCAP/DN08/DNPTRF/DN09/DNLIST
	COMMON /DN10/DNBAPT/DN11/DNFLAB/DN12/DNPTRB
C
	INTEGER	 DNNODE,NODP12,DNSRCE,DNSINK,I2,II2,K2,J2,IHEAD2,
     *	ITAIL2,QHEAD2,QTAIL2,KP2,KSAT2,DNNOP2,LAYMAX,K2DST,K2DP1,MAXD2
	INTEGER	 DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,DNIBIG
	COMMON /DN00/DNARC,DNFVA,DNAUG,DNLFVA,DNLAUG,DNSTGE,
     *		     DNELTM,DNIBIG,DNNODE,DNSRCE,DNSINK,DNNOP2
	EQUIVALENCE (DNFLOW(1),DNTO(1)),(DNFLAB(1),DNDIST(1)),
     *		    (DNFROM(1),DNBADJ(1)),(DNGCAP(1),DNBTOF(1))
	DATA BL3,BL4,BSAT,BCUT,JJT/3Hxxx,4Hyyyy,3HSAT,4H-CUT,1HT/
	DATA JJS/1HS/
C
C***********************************************************************
C
C PREPARES OUTPUT REPORT. MAY ONLY BE CALLED AFTER A CALL TO DNSUB.
C
C***********************************************************************
C
C CALLING CONDITIONS:
C	USER CALLABLE; ONLY AFTER CALLING DNSUB.
C INPUT	ARRAYS:
C	DNFAPT,	DNFADJ,	DNCAP, DNFLOW
C SCRATCH ARRAYS:
C	DNLIST
C OUTPUT ARRAYS:  NONE
C***********************************************************************
C
C PRINT	SUMMARY	SECTION:
	WRITE(IUNREP,1000)
     *	      DNNODE,DNARC,DNSRCE,DNSINK,DNFVA,DNELTM,DNAUG,DNSTGE
C PRINT	CUT-SET	NODES:
	K2=0
	DO 200 I2=1,DNNODE
		IF(DNFAPT(I2).LT.0)GOTO200
		K2=K2+1
		DNLIST(K2)=I2
200	CONTINUE
	WRITE(IUNREP,1003)K2
	WRITE(IUNREP,1004)(DNLIST(I2),I2=1,K2)
	WRITE(IUNREP,1001)
C PRINT	ARC LIST:
	NTOTC4=0
	NSAINC=0
	NSA=0
	NPOSFL=0
	DO 400 I2=1,DNNODE
	   IBEG4=IABS(DNFAPT(I2))
	   IEND4=IABS(DNFAPT(I2+1))-1
	   DO 300 I4=IBEG4,IEND4
		MFLO4=DNFLOW(I4)
		IF(MFLO4.EQ.0)GOTO300
		NPOSFL=NPOSFL+1
		TAG1=BL3
		IF(MFLO4.NE.DNCAP(I4))GOTO270
			TAG1=BSAT
			NSA=NSA+1
270		TAG2=BL4
C		IF(DNFADJ(I4).LT.0)TAG2=BCUT
		IF(DNFADJ(I4).GE.0)GOTO280
			NTOTC4=NTOTC4+MFLO4
			NSAINC=NSAINC+1
280		ISIDE=JJS
		JSIDE=ISIDE
		IF(DNFAPT(I2).LT.0)ISIDE=JJT
		IFAPLU=DNFADJ(I4)
		IFAPLU=IABS(IFAPLU)
		IF(DNFAPT(IFAPLU).LT.0)JSIDE=JJT
		IF(ISIDE.NE.JSIDE)TAG2=BCUT
C
		WRITE(IUNREP,1002)I2,ISIDE,IFAPLU,JSIDE,
     *				MFLO4,DNCAP(I4),TAG1,TAG2
300	   CONTINUE
400	CONTINUE
C PRINT	MAXFLO VALUE AND COMPUTED MINCUT
	WRITE(IUNREP,1006)NPOSFL,NSA,NSAINC
	IF(DNFVA.NE.NTOTC4)WRITE(IUNREP,1007)DNFVA,NTOTC4
	WRITE(IUNREP,1008)
	RETURN
C
C FORMATS:
1000	FORMAT(
     */13H PROBLEM HAS ,I6,7H NODES,,I6,24H ARCS, SOURCE/SINK PAIR ,2I6
     */16H MAX FLOW VALUE=,I16,21H, PC/AT ELAPSED TIME ,F8.2,6H	SECS.
     */21H TOTAL AUGMENTATIONS=,I6,4H IN ,I6,7H	STAGES)
1001	FORMAT(/25H ARCS WITH POSITIVE FLOW
     *//23H	  TAIL	      HEAD,
     *	40H	      FLOW	 CAPACITY    STATUS/)
1002	FORMAT(I10,1X,A1,I10,1X,A1,2I15,4X,A3,A4)
1003	FORMAT(/10H THERE ARE,I8,29H NODES IN SOURCE SIDE OF CUT,,
     *		10H THEY ARE /)
1004	FORMAT(10I6)
1006	FORMAT(/7H NOTES
     */10X,13H 1) THERE	ARE,I6,26H ARCS	WITH POSITIVE FLOW, ,I5,3H OF
     */10X,22H	    THEM SATURATED (,I5,26H OF WHICH ARE IN MIN	CUT).
     */10X,52H 2)   S  /  T   REFER TO SOURCE /	SINK SIDES OF CUT.)
1007	FORMAT(
     * 10X,41H 3) MAX FLOW VALUE FROM SOURCE TO	SINK IS,I14,1H,
     */10X,41H	    BUT, CUT CAPACITY WAS COMPUTED AS  ,I14)
1008	FORMAT(/15H END	OF REPORT.)
	END
